theory CP
imports
  Trs
  Unification
  Renaming
begin

no_notation
  less_multiset (infix "<#" 50)

fun permute_term :: "'v perm \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term" where
  "permute_term p (Var x) = Var (permute_atom p x)" |
  "permute_term p (Fun f ts) = Fun f (map (permute_term p) ts)"

interpretation term_pt: pt permute_term
  by (default) (induct_tac [!] x, auto simp: map_idI)

adhoc_overloading
  PERMUTE permute_term and
  FRESH term_pt.fresh term_pt.fresh_set

interpretation terms_pt: list_pt permute_term ..

adhoc_overloading
  PERMUTE terms_pt.permute_list and
  FRESH terms_pt.fresh

lemma supp_Var:
  fixes x :: "'v :: infinite"
  shows "term_pt.supp (Var x) = {x}"
  using assms and infinite_UNIV
  by (auto simp: term_pt.supp_def swap_atom)

lemma permute_Fun:
  fixes p :: "('v :: infinite) perm"
  shows "p \<bullet> (Fun f ss) = Fun f (p \<bullet> ss)"
  by (induct ss) (simp)+

lemma supp_Fun':
  "term_pt.supp (Fun f ss) = pt.supp (op \<bullet>) ss"
  by (simp only: term_pt.supp_def terms_pt.supp_def permute_Fun) auto

lemma supp_Fun:
  "term_pt.supp (Fun f ss) = (\<Union>s\<in>set ss. term_pt.supp s)"
  by (simp add: supp_Fun')

lemma supp_vars_term_eq:
  fixes t :: "('f, 'v :: infinite) term"
  shows "term_pt.supp t = vars_term t"
  using assms by (induct t) (simp add: supp_Var supp_Fun)+


interpretation subst_conjugate_pt!: fun_pt permute_atom permute_term ..

abbreviation "conjugate_subst \<equiv> subst_conjugate_pt.permute_fun"

lemma eqvt_term_apply_subst_conjugate [eqvt]:
  "p \<bullet> (t \<cdot> \<sigma>) = (p \<bullet> t) \<cdot> (conjugate_subst p \<sigma>)"
  by (induct t) (simp add: subst_conjugate_pt.permute_fun_def)+

definition permute_subst :: "('v :: infinite) perm \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) subst"
where
  "permute_subst \<pi> \<sigma> x = \<pi> \<bullet> \<sigma> x"

interpretation subst_pt!: pt permute_subst
  by (default) (simp add: permute_subst_def [abs_def])+

adhoc_overloading
  PERMUTE "permute_subst" and
  FRESH subst_pt.fresh

fun permute_ctxt :: "'v perm \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f, 'v) ctxt" where
  "permute_ctxt p \<box> = \<box>" |
  "permute_ctxt p (More f ss1 C ss2) =
    More f (map (permute_term p) ss1) (permute_ctxt p C) (map (permute_term p) ss2)"

interpretation ctxt_pt: pt permute_ctxt
  by (default) (induct_tac [!] x, auto simp: map_idI)

adhoc_overloading
  PERMUTE permute_ctxt and
  FRESH ctxt_pt.fresh

lemma supp_Hole:
  "ctxt_pt.supp \<box> = {}"
  by (simp add: ctxt_pt.supp_def)

lemma permute_More:
  fixes p :: "('v :: infinite) perm"
  shows "p \<bullet> (More f ss1 C ss2) = More f (p \<bullet> ss1) (p \<bullet> C) (p \<bullet> ss2)"
  by (simp)

lemma supp_More':
  "ctxt_pt.supp (More f ss1 C ss2) =
    pt.supp (op \<bullet>) ss1 \<union> ctxt_pt.supp C \<union>  pt.supp (op \<bullet>) ss2"
  by (simp only: ctxt_pt.supp_def terms_pt.supp_def) auto

lemma supp_More:
  "ctxt_pt.supp (More f ss1 C ss2) =
    (\<Union>s\<in>set ss1. term_pt.supp s) \<union> ctxt_pt.supp C \<union> (\<Union>s\<in>set ss2. term_pt.supp s)"
  by (simp add: supp_More')

lemma supp_vars_ctxt_eq:
  fixes C :: "('f, 'v :: infinite) ctxt"
  shows "ctxt_pt.supp C = vars_ctxt C"
  by (induct C) (auto simp: supp_Hole supp_More supp_vars_term_eq)

lemma term_apply_ctxt_eqvt [eqvt]:
  fixes p :: "('v :: infinite) perm"
  shows "p \<bullet> (C\<langle>t\<rangle>) = (p \<bullet> C)\<langle>p \<bullet> t\<rangle>"
  by (induct C) simp+

interpretation rule_pt: prod_pt permute_term permute_term ..

adhoc_overloading
  PERMUTE "rule_pt.permute_prod" and
  FRESH rule_pt.fresh rule_pt.fresh_set

lemma supp_vars_rule_eq:
  fixes r :: "('f, 'v :: infinite) rule"
  shows "rule_pt.supp r = vars_term (fst r) \<union> vars_term (snd r)"
  by (cases r) (simp add: supp_vars_term_eq)

lemma rstepD [dest]:
  assumes "(s, t) \<in> rstep R"
  shows "\<exists>l r C \<sigma>. (l, r) \<in> R \<and> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  using rstepE [OF assms] by metis

interpretation trs_pt: set_pt rule_pt.permute_prod ..

adhoc_overloading
  PERMUTE trs_pt.permute_set and
  FRESH trs_pt.fresh

lemma rule_mem_trs_iff [iff]:
  fixes p :: "('v :: infinite) perm"
  shows "(p \<bullet> s, p \<bullet> t) \<in> p \<bullet> R \<longleftrightarrow> (s, t) \<in> R"
  unfolding rule_pt.permute_prod.simps [symmetric] trs_pt.mem_permute_iff ..

lemma inv_rule_mem_trs_simps [simp]:
  fixes p :: "('v :: infinite) perm"
  shows "(-p \<bullet> s, -p \<bullet> t) \<in> R \<longleftrightarrow> (s, t) \<in> p \<bullet> R"
    and "(s, t) \<in> -p \<bullet> R \<longleftrightarrow> (p \<bullet> s, p \<bullet> t) \<in> R"
  unfolding rule_pt.permute_prod.simps [symmetric]
  by (metis trs_pt.inv_mem_simps(1))
     (metis trs_pt.inv_mem_simps(2))

lemma rstep_eqvt [eqvt]:
  fixes p :: "('v :: infinite) perm"
  shows "p \<bullet> rstep R = rstep (p \<bullet> R)"
proof
  show "rstep (p \<bullet> R) \<subseteq> p \<bullet> rstep R"
  proof (rule subrelI)
    fix s t
    assume "(s, t) \<in> rstep (p \<bullet> R)"
    then show "(s, t) \<in> p \<bullet> rstep R"
    proof (induct)
      case (IH C \<sigma> l r)
      have "(l, r) \<in> p \<bullet> R" by fact
      then have "(p \<bullet> (-p) \<bullet> l, p \<bullet> (-p) \<bullet> r) \<in> p \<bullet> R" by simp+
      then have "(-p \<bullet> l, -p \<bullet> r) \<in> R" by simp
      then have "(-p \<bullet> l, -p \<bullet> r) \<in> rstep R" by blast
      then have "((-p \<bullet> C)\<langle>-p \<bullet> l \<cdot> (conjugate_subst (-p) \<sigma>)\<rangle>, (-p \<bullet> C)\<langle>-p \<bullet> r \<cdot> (conjugate_subst (-p) \<sigma>)\<rangle>) \<in> rstep R" by blast
      then have "(p \<bullet> (-p \<bullet> C)\<langle>-p \<bullet> l \<cdot> (conjugate_subst (-p) \<sigma>)\<rangle>, p \<bullet> (-p \<bullet> C)\<langle>-p \<bullet> r \<cdot> (conjugate_subst (-p) \<sigma>)\<rangle>) \<in> p \<bullet> rstep R"
        unfolding rule_mem_trs_iff .
      then show "(C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> p \<bullet> rstep R" by (simp add: eqvt)
    qed
  qed
next
  show "p \<bullet> rstep R \<subseteq> rstep (p \<bullet> R)"
  proof (rule subrelI)
    fix s t
    assume "(s, t) \<in> p \<bullet> rstep R"
    then have "(-p \<bullet> s, -p \<bullet> t) \<in> -p \<bullet> p \<bullet> rstep R" by simp
    then have "(-p \<bullet> s, -p \<bullet> t) \<in> rstep R" by simp
    then show "(s, t) \<in> rstep (p \<bullet> R)"
    proof (induct x\<equiv>"-p \<bullet> s" y\<equiv>"-p \<bullet> t")
      case (IH C \<sigma> l r)
      then have "(p \<bullet> l, p \<bullet> r) \<in> p \<bullet> R"
        and [simp]: "s = p \<bullet> C\<langle>l \<cdot> \<sigma>\<rangle>" "t = p \<bullet> C\<langle>r \<cdot> \<sigma>\<rangle>" by simp+
      then have "((p \<bullet> C)\<langle>(p \<bullet> l) \<cdot> (conjugate_subst p \<sigma>)\<rangle>, (p \<bullet> C)\<langle>(p \<bullet> r) \<cdot> (conjugate_subst p \<sigma>)\<rangle>) \<in> rstep (p \<bullet> R)" by blast
      then show "(s, t) \<in> rstep (p \<bullet> R)" by (simp add: eqvt)
    qed
  qed
qed

lemma term_apply_subst_Var_Rep_perm [simp]:
  "t \<cdot> (Var \<circ> Rep_perm p) = p \<bullet> t"
  by (induct t) (simp add: permute_atom_def)+

text {*The rewrite relation is invariant under variable renamings.*}
lemma rstep_permute [simp]:
  fixes p :: "('v :: infinite) perm"
  shows "rstep (p \<bullet> R) = rstep R"
proof
  show "rstep (p \<bullet> R) \<subseteq> rstep R"
  proof (rule subrelI)
    fix s t
    assume "(s, t) \<in> rstep (p \<bullet> R)"
    then show "(s, t) \<in> rstep R"
    proof (rule rstepE)
      fix l r C \<sigma>
      assume "(l, r) \<in> p \<bullet> R" and [simp]: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      then have "(-p \<bullet> l, -p \<bullet> r) \<in> -p \<bullet> p \<bullet> R" by simp
      then have "(-p \<bullet> l, -p \<bullet> r) \<in> R" by simp
      then have "((-p \<bullet> l) \<cdot> (Var \<circ> Rep_perm p), (-p \<bullet> r) \<cdot> (Var \<circ> (Rep_perm p))) \<in> rstep R" by blast
      then have "(l, r) \<in> rstep R" by simp
      then show "(s, t) \<in> rstep R" by auto
    qed
  qed
next
  show "rstep R \<subseteq> rstep (p \<bullet> R)"
  proof (rule subrelI)
    fix s t
    assume "(s, t) \<in> rstep R"
    then show "(s, t) \<in> rstep (p \<bullet> R)"
    proof (rule rstepE)
      fix l r C \<sigma>
      assume "(l, r) \<in> R" and [simp]: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      then have "(p \<bullet> l, p \<bullet> r) \<in> p \<bullet> R" by simp
      then have "(C\<langle>(p \<bullet> l) \<cdot> (Var \<circ> Rep_perm (-p)) \<cdot> \<sigma>\<rangle>, C\<langle>(p \<bullet> r) \<cdot> (Var \<circ> Rep_perm (-p)) \<cdot> \<sigma>\<rangle>) \<in> rstep (p \<bullet> R)" by blast
      then show "(s, t) \<in> rstep (p \<bullet> R)" by simp
    qed
  qed
qed

lemma rstep_imp_perm_rstep:
  fixes p :: "('v :: infinite) perm"
  shows "(s, t) \<in> rstep R \<Longrightarrow> (p \<bullet> s, p \<bullet> t) \<in> rstep R"
  by (subst (asm) rule_mem_trs_iff [symmetric]) (simp add: eqvt)

lemma perm_rstep_imp_rstep:
  fixes p :: "('v :: infinite) perm"
  shows "(p \<bullet> s, p \<bullet> t) \<in> rstep R \<Longrightarrow> (s, t) \<in> rstep R"
  by (subst (asm) rule_mem_trs_iff [symmetric, of _ _ _ "-p"]) (simp add: eqvt)

lemma rstep_permute_iff [iff]:
  fixes p :: "('v :: infinite) perm"
  shows "(p \<bullet> s, p \<bullet> t) \<in> rstep R \<longleftrightarrow> (s, t) \<in> rstep R"
  by (metis perm_rstep_imp_rstep rstep_imp_perm_rstep)

lemma term_apply_subst_Var_Abs_perm:
  "f \<in> perms \<Longrightarrow> t \<cdot> (Var \<circ> f) = Abs_perm f \<bullet> t"
  by (metis Abs_perm_inverse term_apply_subst_Var_Rep_perm)

lemma finite_term_supp: "finite (term_pt.supp t)"
  unfolding supp_vars_term_eq
  by (induct t) simp+

lemma finite_rule_supp: "finite (rule_pt.supp (l, r))"
  by (simp add: finite_term_supp)

interpretation term_fs: fs permute_term
  by (default) (rule finite_term_supp)

interpretation rule_fs: fs rule_pt.permute_prod
  by (default) (auto simp: finite_rule_supp finite_term_supp)

lemma vars_rule_disjoint:
  fixes l r u v :: "('f, 'v::infinite) term"
  shows "\<exists>p. vars_rule (p \<bullet> (l, r)) \<inter> vars_rule (u, v) = {}"
proof -
  from rule_fs.supp_fresh_set obtain p
    where "rule_pt.supp (p \<bullet> (l, r)) \<sharp> (u, v)" by blast
  from rule_pt.fresh_set_disjoint [OF this]
    show ?thesis
    by (auto simp: supp_vars_rule_eq vars_rule_def)
qed

lemma [termination_simp]:
  "x \<in> set xs \<Longrightarrow> size x < list_size size xs"
  by (induct xs) auto

fun funposs :: "('f, 'v) term \<Rightarrow> pos set" where
  "funposs (Var x) = {}" |
  "funposs (Fun f ts) = {\<epsilon>} \<union> (\<Union>i<length ts. {i <# p | p. p \<in> funposs (ts ! i)})"

lemma funposs_imp_poss:
  assumes "p \<in> funposs t"
  shows "p \<in> poss t"
  using assms by (induct t arbitrary: p) auto

fun varposs :: "('f, 'v) term \<Rightarrow> pos set" where
  "varposs (Var x) = {\<epsilon>}" |
  "varposs (Fun f ts) = (\<Union>i<length ts. {i <# p | p. p \<in> varposs (ts ! i)})"

lemma varposs_imp_poss:
  assumes "p \<in> varposs t"
  shows "p \<in> poss t"
  using assms by (induct t arbitrary: p) auto

lemma poss_simps [symmetric, simp]:
  "poss t = funposs t \<union> varposs t"
  "poss t = varposs t \<union> funposs t"
  "funposs t = poss t - varposs t"
  "varposs t = poss t - funposs t"
  by (induct_tac [!] t) auto

(*TODO: move*)
lemma vars_term_subt_at:
  assumes "p \<in> poss t"
  shows "vars_term (t |_ p) \<subseteq> vars_term t"
  using vars_ctxt_pos_term [OF assms] by simp

definition
  overlap ::
    "('f, 'v::infinite) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow>
    ('f, 'v) rule \<Rightarrow> pos \<Rightarrow> ('f, 'v) rule \<Rightarrow> bool"
where
  "overlap R R' r p r' \<longleftrightarrow>
    (\<exists>p. p \<bullet> r \<in> R) \<and>
    (\<exists>p. p \<bullet> r' \<in> R') \<and>
    vars_rule r \<inter> vars_rule r' = {} \<and>
    p \<in> funposs (fst r') \<and>
    (\<exists>\<sigma>::('f, 'v) subst. fst r \<cdot> \<sigma> = (subt_at (fst r') p) \<cdot> \<sigma>) \<and>
    (p = \<epsilon> \<longrightarrow> \<not> (\<exists>p. p \<bullet> r = r'))"

definition
  cpeaks2 ::
    "('f, 'v :: infinite) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow>
    (('f, 'v) term \<times> ('f, 'v) term \<times> pos \<times> ('f, 'v) term) set"
where
  "cpeaks2 R R' = {(replace_at (fst r') p (snd r) \<cdot> \<sigma>, fst r' \<cdot> \<sigma>, p, snd r' \<cdot> \<sigma>) |
    r p r' \<sigma>. overlap R R' r p r' \<and> \<sigma> = the_mgu (fst r) (subt_at (fst r') p)}"

abbreviation "cpeaks R \<equiv> cpeaks2 R R"

definition
  CP2 :: "('f, 'v::infinite) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) rule set"
where
  "CP2 R R' = {(replace_at (fst r') p (snd r) \<cdot> \<sigma>, snd r' \<cdot> \<sigma>) |
    r p r' \<sigma>. overlap R R' r p r' \<and> \<sigma> = the_mgu (fst r) (subt_at (fst r') p)}"

abbreviation "CP R \<equiv> CP2 R R"

lemma CP2_I:
  assumes "overlap R R' r p r'" and "\<sigma> = the_mgu (fst r) (subt_at (fst r') p)"
  shows "(replace_at (fst r') p (snd r) \<cdot> \<sigma>, snd r' \<cdot> \<sigma>) \<in> CP2 R R'"
  using assms by (force simp: CP2_def)

lemma cpeaks2_I:
  assumes "overlap R R' r p r'" and "\<sigma> = the_mgu (fst r) (subt_at (fst r') p)"
  shows "(replace_at (fst r') p (snd r) \<cdot> \<sigma>, fst r' \<cdot> \<sigma>, p, snd r' \<cdot> \<sigma>) \<in> cpeaks2 R R'"
  using assms by (force simp: cpeaks2_def)

(*TODO: move*)
lemma pos_prefix_same [simp]:
  "pos_prefix p p = Some \<epsilon>"
  by (induct p) simp_all

definition "pos_diff p q = the (pos_prefix q p)"

lemma prefix_pos_diff [simp]:
  assumes "p \<le> q"
  shows "p <#> pos_diff q p = q"
  using prefix_exists [OF assms] by (auto simp: pos_diff_def)

lemma vars_term_eqvt [eqvt]:
  "\<pi> \<bullet> vars_term t = vars_term (\<pi> \<bullet> t)"
  by (simp add: supp_vars_term_eq [symmetric] eqvt)

lemma is_imgu_imp_is_mgu:
  assumes "is_imgu \<sigma> E"
  shows "is_mgu \<sigma> E"
  using assms by (auto simp: is_imgu_def is_mgu_def)

(*TODO: move*)
lemma coincidence_lemma':
  assumes "vars_term t \<subseteq> V"
  shows "t \<cdot> (\<sigma> |s V) = t \<cdot> \<sigma>"
  using assms by (metis in_mono in_subst_restrict term_subst_eq)

lemma permute_term_subst_apply_term:
  "(\<pi> \<bullet> t) \<cdot> \<sigma> = t \<cdot> (\<sigma> \<circ> Rep_perm \<pi>)"
  by (induct t) (simp_all add: permute_atom_def)

lemma permute_subst_subst_compose:
  "\<pi> \<bullet> \<sigma> = \<sigma> \<circ>s (Var \<circ> Rep_perm \<pi>)"
  by (rule ext) (simp add: subst_compose_def permute_subst_def)

lemma varposs_iff:
  "p \<in> varposs t \<longleftrightarrow> (\<exists>x. p \<in> poss t \<and> t |_ p = Var x)"
  by (induct t arbitrary: p) auto

lemma poss_subst_apply_term:
  assumes "p \<in> poss (t \<cdot> \<sigma>)" and "p \<notin> funposs t"
  obtains q r x where "p = q <#> r" and "q \<in> poss t" and "t |_ q = Var x" and "r \<in> poss (\<sigma> x)"
using assms
proof (induct t arbitrary: p)
  case (Fun f ts)
  then show ?case by (auto) (metis append_PCons nth_mem subt_at.simps(2))
qed simp

lemma replace_at_subst_rsteps:
  fixes \<sigma> \<tau> :: "('f, 'v) subst"
  assumes *: "\<And>x. (\<sigma> x, \<tau> x) \<in> (rstep R)\<^sup>*"
    and "p \<in> poss t"
    and "t |_ p = Var x"
  shows "(replace_at (t \<cdot> \<sigma>) p (\<tau> x), t \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
using assms(2-)
proof (induction t arbitrary: p)
  case (Fun f ts)
  then obtain i q where [simp]: "p = i <# q" and i: "i < length ts"
    and q: "q \<in> poss (ts ! i)" and [simp]: "ts ! i |_ q = Var x" by (cases p) auto
  let ?C = "ctxt_of_pos_term q (ts ! i \<cdot> \<sigma>)"
  let ?ts = "map (\<lambda>t. t \<cdot> \<tau>) ts"
  let ?ss = "take i (map (\<lambda>t. t \<cdot> \<sigma>) ts) @ ?C\<langle>\<tau> x\<rangle> # drop (Suc i) (map (\<lambda>t. t \<cdot> \<sigma>) ts)"
  have "\<forall>j<length ts. (?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*"
  proof (intro allI impI)
    fix j
    assume j: "j < length ts"
    moreover
    { assume [simp]: "j = i"
      have "?ss ! j = ?C\<langle>\<tau> x\<rangle>" using i by (simp add: nth_append_take)
      with Fun.IH [of "ts ! i" q]
      have "(?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*" using q and i by simp }
    moreover
    { assume "j < i"
      with i have "?ss ! j = ts ! j \<cdot> \<sigma>"
        and "?ts ! j = ts ! j \<cdot> \<tau>" by (simp_all add: nth_append_take_is_nth_conv)
      then have "(?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*" by (auto simp: * subst_rsteps_imp_rsteps) }
    moreover
    { assume "j > i"
      with i and j have "?ss ! j = ts ! j \<cdot> \<sigma>"
        and "?ts ! j = ts ! j \<cdot> \<tau>" by (simp_all add: nth_append_drop_is_nth_conv)
      then have "(?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*" by (auto simp: * subst_rsteps_imp_rsteps) }
    ultimately show "(?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*" by arith
  qed
  moreover have "i < length ts" by fact
  ultimately show ?case
    by (simp add: args_rsteps_imp_rsteps)
qed simp

(*TODO: move*)
lemma is_mguI:
  fixes \<sigma> :: "('f, 'v) subst"
  assumes "\<forall>(s, t)\<in>E. s \<cdot> \<sigma> = t \<cdot> \<sigma>"
    and "\<And>\<tau> :: ('f, 'v) subst. \<forall>(s, t)\<in>E. s \<cdot> \<tau> = t \<cdot> \<tau> \<Longrightarrow> \<exists>\<gamma> :: ('f, 'v) subst. \<tau> = \<sigma> \<circ>s \<gamma>"
  shows "is_mgu \<sigma> E"
  using assms by (fastforce simp: is_mgu_def unifiers_def)

(*TODO: move*)
lemma replace_at_ident [simp]:
  assumes "p \<in> poss t" and "t |_ p = s"
  shows "replace_at t p s = t"
  using assms by (metis ctxt_supt_id)

lemma substs_rsteps:
  assumes "\<And>x. (\<sigma> x, \<tau> x) \<in> (rstep R)\<^sup>*"
  shows "(t \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
proof (induct t)
  case (Var y)
  show ?case using assms by simp_all
next
  case (Fun f ts)
  then have "\<forall>i<length (map (\<lambda>t. t \<cdot> \<sigma>) ts).
    (map (\<lambda>t. t \<cdot> \<sigma> ) ts ! i, map (\<lambda>t. t \<cdot> \<tau>) ts ! i) \<in> (rstep R)\<^sup>*" by auto
  from args_rsteps_imp_rsteps [OF _ this] show ?case by simp
qed

lemma CP2_cpeaks2_conv:
  "CP2 R R' = (\<lambda>(t, s, p, u). (t, u)) ` cpeaks2 R R'"
  by (force simp: CP2_def cpeaks2_def)

inductive_set CS3 for A
where
  "(t, s, p, u) \<in> A \<Longrightarrow> (C\<langle>t \<cdot> \<sigma>\<rangle>, C\<langle>s \<cdot> \<sigma>\<rangle>, hole_pos C <#> p, C\<langle>u \<cdot> \<sigma>\<rangle>) \<in> CS3 A"

inductive_set S3 for A
where
  subst: "(t, s, p, u) \<in> A \<Longrightarrow> (t \<cdot> \<sigma>, s \<cdot> \<sigma>, p, u \<cdot> \<sigma>) \<in> S3 A"

lemma CS3_ctxt:
  assumes "(t, s, p, u) \<in> CS3 A"
  shows "(C\<langle>t\<rangle>, C\<langle>s\<rangle>, hole_pos C <#> p, C\<langle>u\<rangle>) \<in> CS3 A"
proof -
  from assms obtain D \<sigma> t' s' u' p'
    where "(t', s', p', u') \<in> A" and "t = D\<langle>t' \<cdot> \<sigma>\<rangle>" "s = D\<langle>s' \<cdot> \<sigma>\<rangle>" "u = D\<langle>u' \<cdot> \<sigma>\<rangle>"
    and "p = hole_pos D <#> p'"
    by (cases)
  with CS3.intros [OF this(1), of "C \<circ>c D" \<sigma>] show ?thesis by simp
qed

inductive_set rstep_pos :: "('f, 'v) trs \<Rightarrow> pos \<Rightarrow> ('f, 'v) term rel" for R and p
where
  rule [intro]: "(l, r) \<in> R \<Longrightarrow> p \<in> poss s \<Longrightarrow> s |_ p = l \<cdot> \<sigma> \<Longrightarrow>
    (s, replace_at s p (r \<cdot> \<sigma>)) \<in> rstep_pos R p"

lemma rstep_pos_rstep_r_p_s_conv:
  "rstep_pos R p = {(s, t) | s t r \<sigma>. (s, t) \<in> rstep_r_p_s R r p \<sigma>}"
  apply (auto simp: rstep_r_p_s_def Let_def elim!: rstep_pos.cases)
  apply (metis replace_at_ident)
  by (metis replace_at_subt_at rstep_pos.intros)

lemma rstep_rstep_pos_conv:
  "rstep R = {(s, t) | s t p. (s, t) \<in> rstep_pos R p}"
  by (auto simp: rstep_pos_rstep_r_p_s_conv)
     (metis rstep_iff_rstep_r_p_s)+

text {*Special case: A peak where one step is above the other is either joinable or originates
from a critical pair.*}
lemma peak_above_imp_join_or_cpeaks:
  fixes R :: "('f, 'v :: infinite) trs"
    and \<sigma>\<^sub>1 \<sigma>\<^sub>2 :: "('f, 'v) subst"
  assumes rule_variants: "\<exists>\<pi>. \<pi> \<bullet> (l\<^sub>1, r\<^sub>1) \<in> R" "\<exists>\<pi>. \<pi> \<bullet> (l\<^sub>2, r\<^sub>2) \<in> R"
    and disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_rule (l\<^sub>2, r\<^sub>2) = {}"
    and "p\<^sub>2 \<le> p\<^sub>1"

    and "vars_term r\<^sub>1 \<subseteq> vars_term l\<^sub>1"
    and p\<^sub>1: "p\<^sub>1 \<in> poss s"
    and sp\<^sub>1: "s |_ p\<^sub>1 = l\<^sub>1 \<cdot> \<sigma>\<^sub>1"
    and t: "t = replace_at s p\<^sub>1 (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)"

    and p\<^sub>2: "p\<^sub>2 \<in> poss s"
    and sp\<^sub>2: "s |_ p\<^sub>2 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2"
    and u: "u = replace_at s p\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)"

  shows "(t, u) \<in> (rstep R)\<^sup>\<down> \<or>
    (\<exists>v w. p\<^sub>2 \<in> poss s \<and> t = (ctxt_of_pos_term p\<^sub>2 s)\<langle>v\<rangle> \<and> u = (ctxt_of_pos_term p\<^sub>2 s)\<langle>w\<rangle> \<and>
    (v, s |_ p\<^sub>2, pos_diff p\<^sub>1 p\<^sub>2, w) \<in> S3 (cpeaks R))"
proof -
  def p \<equiv> "pos_diff p\<^sub>1 p\<^sub>2"
  have p\<^sub>1_eq: "p\<^sub>1 = p\<^sub>2 <#> p" using `p\<^sub>2 \<le> p\<^sub>1` by (simp add: p_def)
  then have t: "t = replace_at s p\<^sub>2 (replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) p (r\<^sub>1 \<cdot> \<sigma>\<^sub>1))"
    by (simp add: t ctxt_of_pos_term_append p\<^sub>2 sp\<^sub>2)
  show ?thesis
  proof (cases "p \<in> funposs l\<^sub>2 \<and> (p = \<epsilon> \<longrightarrow> \<not> (\<exists>\<pi>. \<pi> \<bullet> (l\<^sub>1, r\<^sub>1) = (l\<^sub>2, r\<^sub>2)))")
    case True
    def \<sigma>' \<equiv> "\<lambda>x. if x \<in> vars_rule (l\<^sub>1, r\<^sub>1) then \<sigma>\<^sub>1 x else \<sigma>\<^sub>2 x"

    have p: "p \<in> funposs l\<^sub>2" and "p = \<epsilon> \<Longrightarrow> \<not> (\<exists>\<pi>. \<pi> \<bullet> (l\<^sub>1, r\<^sub>1) = (l\<^sub>2, r\<^sub>2))" using True by auto
    moreover
    have unif: "l\<^sub>1 \<cdot> \<sigma>' = (l\<^sub>2 |_ p) \<cdot> \<sigma>'"
    proof -
      note coinc = coincidence_lemma' [of l\<^sub>1 "vars_rule (l\<^sub>1, r\<^sub>1)"]
      have disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_term (l\<^sub>2 |_ p) = {}"
        using vars_term_subt_at [OF funposs_imp_poss [OF p]] and disj
        by (auto simp: vars_rule_def)
      have "l\<^sub>1 \<cdot> \<sigma>' = l\<^sub>1 \<cdot> (\<sigma>' |s vars_rule (l\<^sub>1, r\<^sub>1))"
        using coinc by (simp add: vars_rule_def)
      also have "\<dots> = l\<^sub>1 \<cdot> (\<sigma>\<^sub>1 |s vars_rule (l\<^sub>1, r\<^sub>1))" by (simp add: \<sigma>'_def)
      also have "\<dots> = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using coinc by (simp add: vars_rule_def)
      also have "\<dots> = (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) |_ p" using subt_at_append [OF p\<^sub>2] and sp\<^sub>1 by (simp add: p\<^sub>1_eq sp\<^sub>2)
      also have "\<dots> = (l\<^sub>2 |_ p) \<cdot> \<sigma>\<^sub>2" using funposs_imp_poss [OF p] by simp
      also have "\<dots> = (l\<^sub>2 |_ p) \<cdot> (\<sigma>\<^sub>2 |s vars_term (l\<^sub>2 |_ p))"
        by (simp add: coincidence_lemma [symmetric])
      also have "\<dots> = (l\<^sub>2 |_ p) \<cdot> (\<sigma>' |s vars_term (l\<^sub>2 |_ p))" using disj by (simp add: \<sigma>'_def)
      finally show ?thesis by (simp add: coincidence_lemma [symmetric])
    qed
    ultimately
    have overlap: "overlap R R (l\<^sub>1, r\<^sub>1) p (l\<^sub>2, r\<^sub>2)"
      using disj and rule_variants by (auto simp: overlap_def)

    def \<sigma> \<equiv> "the_mgu l\<^sub>1 (l\<^sub>2 |_ p)"
    have cpeaks: "(replace_at l\<^sub>2 p r\<^sub>1 \<cdot> \<sigma>, l\<^sub>2 \<cdot> \<sigma>, p, r\<^sub>2 \<cdot> \<sigma>) \<in> cpeaks R"
      using cpeaks2_I [OF overlap refl] by (simp add: \<sigma>_def)

    have "is_mgu \<sigma> {(l\<^sub>1, l\<^sub>2 |_ p)}" by (rule is_mguI) (insert unif the_mgu, auto simp: \<sigma>_def)
    with unif obtain \<tau> where \<sigma>': "\<sigma>' = \<sigma> \<circ>s \<tau>" by (auto simp: is_mgu_def unifiers_def)
    have "(replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) p (r\<^sub>1 \<cdot> \<sigma>\<^sub>1), l\<^sub>2 \<cdot> \<sigma>\<^sub>2, p, r\<^sub>2 \<cdot> \<sigma>\<^sub>2) \<in> S3 (cpeaks R)"
    proof -
      have "l\<^sub>2 \<cdot> \<sigma>\<^sub>2 = l\<^sub>2 \<cdot> \<sigma>'"
      proof -
        have disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_term l\<^sub>2 = {}" using disj by (auto simp: vars_rule_def)
        have "l\<^sub>2 \<cdot> \<sigma>' = l\<^sub>2 \<cdot> (\<sigma>' |s vars_term l\<^sub>2)" by (rule coincidence_lemma)
        also have "\<dots> = l\<^sub>2 \<cdot> (\<sigma>\<^sub>2 |s vars_term l\<^sub>2)" using disj by (simp add: \<sigma>'_def)
        also have "\<dots> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" by (simp add: coincidence_lemma [symmetric])
        finally show ?thesis ..
      qed
      moreover
      have "r\<^sub>1 \<cdot> \<sigma>\<^sub>1 = r\<^sub>1 \<cdot> \<sigma>'"
      proof -
        note coinc = coincidence_lemma' [of r\<^sub>1 "vars_rule (l\<^sub>1, r\<^sub>1)"]
        have "r\<^sub>1 \<cdot> \<sigma>' = r\<^sub>1 \<cdot> (\<sigma>' |s vars_rule (l\<^sub>1, r\<^sub>1))"
          using coinc by (simp add: vars_rule_def)
        also have "\<dots> = r\<^sub>1 \<cdot> (\<sigma>\<^sub>1 |s vars_rule (l\<^sub>1, r\<^sub>1))" by (simp add: \<sigma>'_def)
        also have "\<dots> = r\<^sub>1 \<cdot> \<sigma>\<^sub>1" using coinc by (simp add: vars_rule_def)
        finally show ?thesis ..
      qed
      moreover
      have "r\<^sub>2 \<cdot> \<sigma>\<^sub>2 = r\<^sub>2 \<cdot> \<sigma>'"
      proof -
        have disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_term r\<^sub>2 = {}" using disj by (auto simp: vars_rule_def)
        have "r\<^sub>2 \<cdot> \<sigma>' = r\<^sub>2 \<cdot> (\<sigma>' |s vars_term r\<^sub>2)" by (rule coincidence_lemma)
        also have "\<dots> = r\<^sub>2 \<cdot> (\<sigma>\<^sub>2 |s vars_term r\<^sub>2)" using disj by (simp add: \<sigma>'_def)
        also have "\<dots> = r\<^sub>2 \<cdot> \<sigma>\<^sub>2" by (simp add: coincidence_lemma [symmetric])
        finally show ?thesis ..
      qed
      ultimately
      show ?thesis
        using S3.intros [OF cpeaks, of \<tau>]
        and funposs_imp_poss [OF p]
        by (simp add: \<sigma>' ctxt_of_pos_term_subst)
    qed
    moreover have "p\<^sub>2 \<in> poss s" by fact
    ultimately show ?thesis using p\<^sub>2 and sp\<^sub>2 by (auto simp: t u p_def)
  next
    case False
    then have "(p = \<epsilon> \<and> (\<exists>\<pi>. \<pi> \<bullet> (l\<^sub>1, r\<^sub>1) = (l\<^sub>2, r\<^sub>2))) \<or> p \<notin> funposs l\<^sub>2" by blast
    then show ?thesis
    proof (elim disjE conjE exE)
      txt {*The variable condition is only needed here.*}
      fix \<pi>
      assume [simp]: "p = \<epsilon>" and "\<pi> \<bullet> (l\<^sub>1, r\<^sub>1) = (l\<^sub>2, r\<^sub>2)"
      then have "p\<^sub>2 = p\<^sub>1" and "l\<^sub>2 = \<pi> \<bullet> l\<^sub>1" and [simp]: "r\<^sub>2 = \<pi> \<bullet> r\<^sub>1" by (simp_all add: p\<^sub>1_eq eqvt)
      then have "l\<^sub>1 \<cdot> \<sigma>\<^sub>1 = l\<^sub>1 \<cdot> (\<sigma>\<^sub>2 \<circ> Rep_perm \<pi>)"
        using sp\<^sub>1 and sp\<^sub>2 by (simp add: permute_term_subst_apply_term)
      moreover
      have "vars_term r\<^sub>1 \<subseteq> vars_term l\<^sub>1" by fact
      ultimately
      have "r\<^sub>1 \<cdot> \<sigma>\<^sub>1 = r\<^sub>1 \<cdot> (\<sigma>\<^sub>2 \<circ> Rep_perm \<pi>)" using vars_term_subset_subst_eq by blast
      then have "r\<^sub>1 \<cdot> \<sigma>\<^sub>1 = \<pi> \<bullet> r\<^sub>1 \<cdot> \<sigma>\<^sub>2" by (simp add: permute_term_subst_apply_term)
      then show ?thesis by (auto simp: t u)
    next
      txt {*Variable overlap.*}
      have step1: "(l\<^sub>1, r\<^sub>1) \<in> rstep R" and step2: "(l\<^sub>2, r\<^sub>2) \<in> rstep R"
        using rule_variants by (auto simp: eqvt) (metis rstep.id rstep_permute_iff)+
      have "p \<in> poss (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)" by (metis p\<^sub>1_eq p\<^sub>1 poss_append_poss sp\<^sub>2)
      moreover
      assume "p \<notin> funposs l\<^sub>2"
      ultimately
      obtain q\<^sub>1 q\<^sub>2 x
        where [simp]: "p = q\<^sub>1 <#> q\<^sub>2" and q\<^sub>1: "q\<^sub>1 \<in> poss l\<^sub>2"
        and l\<^sub>2q\<^sub>1: "l\<^sub>2 |_ q\<^sub>1 = Var x" and q\<^sub>2: "q\<^sub>2 \<in> poss (\<sigma>\<^sub>2 x)"
        by (rule poss_subst_apply_term)
      moreover
      have [simp]: "l\<^sub>2 \<cdot> \<sigma>\<^sub>2 |_ p = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" by (metis p\<^sub>1_eq p\<^sub>2 sp\<^sub>1 sp\<^sub>2 subt_at_append)
      ultimately
      have [simp]: "\<sigma>\<^sub>2 x |_ q\<^sub>2 = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" by simp

      def \<sigma>\<^sub>2' \<equiv> "\<lambda>y. if y = x then replace_at (\<sigma>\<^sub>2 x) q\<^sub>2 (r\<^sub>1 \<cdot> \<sigma>\<^sub>1) else \<sigma>\<^sub>2 y"

      have "(\<sigma>\<^sub>2 x, \<sigma>\<^sub>2' x) \<in> rstep R"
      proof -
        let ?C = "ctxt_of_pos_term q\<^sub>2 (\<sigma>\<^sub>2 x)"
        have "(?C\<langle>l\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>, ?C\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>) \<in> rstep R" using step1 by blast
        then show ?thesis using q\<^sub>2 by (simp add: \<sigma>\<^sub>2'_def)
      qed
      then have *: "\<And>x. (\<sigma>\<^sub>2 x, \<sigma>\<^sub>2' x) \<in> (rstep R)\<^sup>*" by (auto simp: \<sigma>\<^sub>2'_def)
      then have "(r\<^sub>2 \<cdot> \<sigma>\<^sub>2, r\<^sub>2 \<cdot> \<sigma>\<^sub>2') \<in> (rstep R)\<^sup>*" by (rule substs_rsteps)
      then have "(u, replace_at s p\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2')) \<in> (rstep R)\<^sup>*" by (auto simp: u rsteps_closed_ctxt)
      moreover
      have "(t, replace_at s p\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2')) \<in> (rstep R)\<^sup>*"
      proof -
        have "replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) p (r\<^sub>1 \<cdot> \<sigma>\<^sub>1) = replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) q\<^sub>1 (\<sigma>\<^sub>2' x)"
          using q\<^sub>1 and q\<^sub>2 by (simp add: \<sigma>\<^sub>2'_def ctxt_of_pos_term_append l\<^sub>2q\<^sub>1)
        moreover
        have "(replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) q\<^sub>1 (\<sigma>\<^sub>2' x), l\<^sub>2 \<cdot> \<sigma>\<^sub>2') \<in> (rstep R)\<^sup>*"
          by (rule replace_at_subst_rsteps [OF * q\<^sub>1 l\<^sub>2q\<^sub>1])
        moreover
        have "(l\<^sub>2 \<cdot> \<sigma>\<^sub>2', r\<^sub>2 \<cdot> \<sigma>\<^sub>2') \<in> rstep R" using step2 by blast
        ultimately
        have "(replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) p (r\<^sub>1 \<cdot> \<sigma>\<^sub>1), r\<^sub>2 \<cdot> \<sigma>\<^sub>2') \<in> (rstep R)\<^sup>*" by auto
        then show ?thesis by (auto simp: p\<^sub>2 t rsteps_closed_ctxt)
      qed
      ultimately show ?thesis by blast
    qed
  qed
qed

text {*Peaks are either joinable or originate from a critical peak.*}
(*Lemma 5.1.9*)
lemma peak_imp_join_or_S3_cpeaks:
  fixes R :: "('f, 'v :: infinite) trs"
  assumes variable_condition: "\<forall>(l, r)\<in>R. vars_term r \<subseteq> vars_term l"
    and peak: "(s, t) \<in> rstep_pos R p\<^sub>1" "(s, u) \<in> rstep_pos R p\<^sub>2"
  shows "(t, u) \<in> (rstep R)\<^sup>\<down> \<or> (\<exists>v w.
    (p\<^sub>2 \<le> p\<^sub>1 \<and> t = (ctxt_of_pos_term p\<^sub>2 s)\<langle>v\<rangle> \<and> u = (ctxt_of_pos_term p\<^sub>2 s)\<langle>w\<rangle> \<and> (v, s |_ p\<^sub>2, pos_diff p\<^sub>1 p\<^sub>2, w) \<in> S3 (cpeaks R)) \<or>
    (p\<^sub>1 \<le> p\<^sub>2 \<and> t = (ctxt_of_pos_term p\<^sub>1 s)\<langle>v\<rangle> \<and> u = (ctxt_of_pos_term p\<^sub>1 s)\<langle>w\<rangle> \<and> (w, s |_ p\<^sub>1, pos_diff p\<^sub>2 p\<^sub>1, v) \<in> S3 (cpeaks R)))"
proof -
  obtain l\<^sub>1' r\<^sub>1' \<sigma>\<^sub>1' l\<^sub>2 r\<^sub>2 \<sigma>\<^sub>2
    where rule1': "(l\<^sub>1', r\<^sub>1') \<in> R" and rule2: "(l\<^sub>2, r\<^sub>2) \<in> R"
    and p\<^sub>1: "p\<^sub>1 \<in> poss s" and p\<^sub>2: "p\<^sub>2 \<in> poss s"
    and s1': "replace_at s p\<^sub>1 (l\<^sub>1' \<cdot> \<sigma>\<^sub>1') = s" and t': "t = replace_at s p\<^sub>1 (r\<^sub>1' \<cdot> \<sigma>\<^sub>1')"
    and s2: "replace_at s p\<^sub>2 (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) = s" and u: "u = replace_at s p\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)"
    using rstep_pos.cases [OF peak(1)]
      and rstep_pos.cases [OF peak(2)]
      by (metis ctxt_supt_id)

  from vars_rule_disjoint obtain \<pi>\<^sub>1
    where \<pi>\<^sub>1: "vars_rule (\<pi>\<^sub>1 \<bullet> (l\<^sub>1', r\<^sub>1')) \<inter> vars_rule (l\<^sub>2, r\<^sub>2) = {}" ..
  def l\<^sub>1 \<equiv> "\<pi>\<^sub>1 \<bullet> l\<^sub>1'" and r\<^sub>1 \<equiv> "\<pi>\<^sub>1 \<bullet> r\<^sub>1'" and \<sigma>\<^sub>1 \<equiv> "(Var \<circ> Rep_perm (-\<pi>\<^sub>1)) \<circ>s \<sigma>\<^sub>1'"
  note rename = l\<^sub>1_def r\<^sub>1_def \<sigma>\<^sub>1_def

  have disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_rule (l\<^sub>2, r\<^sub>2) = {}" using \<pi>\<^sub>1 by (auto simp: eqvt rename)
  have s1: "replace_at s p\<^sub>1 (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) = s" and t: "t = replace_at s p\<^sub>1 (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)"
    by (simp_all add: s1' t' rename)

  have "-\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> R" and "0 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> R"
    using rule1' and rule2 by (simp_all add: rename)
  then have rule_variants: "\<exists>\<pi>. \<pi> \<bullet> (l\<^sub>1, r\<^sub>1) \<in> R" "\<exists>\<pi>. \<pi> \<bullet> (l\<^sub>2, r\<^sub>2) \<in> R" by blast+

  show ?thesis
  proof (cases "p\<^sub>1 \<bottom> p\<^sub>2")
    assume "p\<^sub>1 \<bottom> p\<^sub>2"
    moreover
    have "(s, t) \<in> rstep_r_p_s R (l\<^sub>1', r\<^sub>1') p\<^sub>1 \<sigma>\<^sub>1'"
      by (simp add: p\<^sub>1 rstep_r_p_s_def s1' t' rule1')
    moreover
    have "(s, u) \<in> rstep_r_p_s R (l\<^sub>2, r\<^sub>2) p\<^sub>2 \<sigma>\<^sub>2"
      by (simp add: p\<^sub>2 rstep_r_p_s_def s2 u rule2)
    ultimately
    obtain v where "(t, v) \<in> rstep R" and "(u, v) \<in> rstep R"
      by (blast dest: parallel_steps rstep_r_p_s_imp_rstep)
    then show ?thesis by blast
  next
    have sp\<^sub>1: "s |_ p\<^sub>1 = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using s1 and p\<^sub>1 by (metis replace_at_subt_at)
    have sp\<^sub>2: "s |_ p\<^sub>2 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" using s2 and p\<^sub>2 by (metis replace_at_subt_at)

    assume "\<not> (p\<^sub>1 \<bottom> p\<^sub>2)"
    then have "p\<^sub>1 \<le> p\<^sub>2 \<or> p\<^sub>2 \<le> p\<^sub>1" by (metis parallel_pos)
    then show ?thesis
    proof
      have vars: "vars_term r\<^sub>1 \<subseteq> vars_term l\<^sub>1"
        using variable_condition [THEN bspec, OF rule1']
        and atom_set_pt.permute_set_subset [of "vars_term r\<^sub>1'" "vars_term l\<^sub>1'"]
        by (auto simp add: rename supp_vars_term_eq [symmetric] term_fs.supp_eqvt [symmetric])
      assume "p\<^sub>2 \<le> p\<^sub>1"
      with peak_above_imp_join_or_cpeaks [OF rule_variants disj this vars p\<^sub>1 sp\<^sub>1 t p\<^sub>2 sp\<^sub>2 u]
        show ?thesis by (auto simp: s2)
    next
      have vars: "vars_term r\<^sub>2 \<subseteq> vars_term l\<^sub>2"
        using variable_condition [THEN bspec, OF rule2] by blast
      have disj: "vars_rule (l\<^sub>2, r\<^sub>2) \<inter> vars_rule (l\<^sub>1, r\<^sub>1) = {}" using disj by blast
      assume "p\<^sub>1 \<le> p\<^sub>2"
      with peak_above_imp_join_or_cpeaks [OF rule_variants(2, 1) disj this vars p\<^sub>2 sp\<^sub>2 u p\<^sub>1 sp\<^sub>1 t]
        show ?thesis by (auto simp: s1)
    qed
  qed
qed

lemma CS3_imp_rstep:
  assumes "(t, s, p, u) \<in> CS3 A"
  shows "(t, u) \<in> rstep ((\<lambda>(w, x, y, z). (w, z)) ` A)"
  using assms by (cases) force

lemma S3_imp_CS3:
  assumes "p \<in> poss s"
    and S3: "(v, s |_ p, q, w) \<in> S3 A"
  shows "((ctxt_of_pos_term p s)\<langle>v\<rangle>, s, p <#> q, (ctxt_of_pos_term p s)\<langle>w\<rangle>) \<in> CS3 A"
using S3
proof (cases)
  case (subst t' s' u' \<sigma>)
  with CS3.intros [OF subst(4), of "ctxt_of_pos_term p s" \<sigma>]
    show ?thesis by (simp add: assms)
qed

lemma peak_imp_join_or_CS3_cpeaks:
  fixes R :: "('f, 'v :: infinite) trs"
  assumes vc: "\<forall>(l, r)\<in>R. vars_term r \<subseteq> vars_term l"
    and peak: "(s, t) \<in> rstep R" "(s, u) \<in> rstep R"
  shows "(t, u) \<in> (rstep R)\<^sup>\<down> \<or> (\<exists>p. (t, s, p, u) \<in> CS3 (cpeaks R) \<or> (u, s, p, t) \<in> CS3 (cpeaks R))"
proof -
  from peak obtain p q
    where peak: "(s, t) \<in> rstep_pos R p" "(s, u) \<in> rstep_pos R q"
    by (auto simp: rstep_rstep_pos_conv)
  then have "p \<in> poss s" and "q \<in> poss s" by (auto elim: rstep_pos.cases)
  with peak_imp_join_or_S3_cpeaks [OF vc peak]
    show ?thesis by (auto dest: S3_imp_CS3)
qed

lemma peak_imp_join_or_CP:
  fixes R :: "('f, 'v :: infinite) trs"
  assumes "\<forall>(l, r)\<in>R. vars_term r \<subseteq> vars_term l"
    and "(s, t) \<in> rstep R" "(s, u) \<in> rstep R"
  shows "(t, u) \<in> (rstep R)\<^sup>\<down> \<or> (t, u) \<in> (rstep (CP R))\<^sup>\<leftrightarrow>"
  using peak_imp_join_or_CS3_cpeaks [OF assms]
  by (auto simp: CP2_cpeaks2_conv dest: CS3_imp_rstep simp del: reverseTrs)

lemma join_ctxt [intro]:
  assumes "(s, t) \<in> (rstep R)\<^sup>\<down>"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> (rstep R)\<^sup>\<down>"
proof -
  from assms obtain u where "(s, u) \<in> (rstep R)\<^sup>*" and "(t, u) \<in> (rstep R)\<^sup>*" by auto
  then have "(C\<langle>s\<rangle>, C\<langle>u\<rangle>) \<in> (rstep R)\<^sup>*" and "(C\<langle>t\<rangle>, C\<langle>u\<rangle>) \<in> (rstep R)\<^sup>*" by (auto intro: rsteps_closed_ctxt)
  then show ?thesis by blast
qed

(*TODO: move*)
lemma join_subst [intro]:
  assumes "(s, t) \<in> (rstep R)\<^sup>\<down>"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep R)\<^sup>\<down>"
proof -
  from assms obtain u where "(s, u) \<in> (rstep R)\<^sup>*" and "(t, u) \<in> (rstep R)\<^sup>*" by auto
  then have "(s \<cdot> \<sigma>, u \<cdot> \<sigma>) \<in> (rstep R)\<^sup>*" and "(t \<cdot> \<sigma>, u \<cdot> \<sigma>) \<in> (rstep R)\<^sup>*" by (auto intro: rsteps_closed_subst)
  then show ?thesis by blast
qed

(*TODO: move*)
lemma join_sym:
  "(s, t) \<in> A\<^sup>\<down> \<Longrightarrow> (t, s) \<in> A\<^sup>\<down>"
  by auto

lemma CP_join_imp_peak_join:
  assumes "\<forall>(l, r)\<in>R. vars_term r \<subseteq> vars_term l"
    and *: "\<forall>(s, t)\<in>CP R. (s, t) \<in> (rstep R)\<^sup>\<down>"
  shows "(rstep R)\<inverse> O (rstep R) \<subseteq> (rstep R)\<^sup>\<down>"
proof -
  have "(rstep (CP R))\<^sup>\<leftrightarrow> \<subseteq> (rstep R)\<^sup>\<down>"
  proof (intro subrelI)
    fix s t
    assume "(s, t) \<in> (rstep (CP R))\<^sup>\<leftrightarrow>"
    then have "(s, t) \<in> rstep (CP R) \<or> (t, s) \<in> rstep (CP R)" by auto
    then show "(s, t) \<in> (rstep R)\<^sup>\<down>"
    proof
      assume "(s, t) \<in> rstep (CP R)"
      then obtain C \<sigma> u v where cp: "(u, v) \<in> CP R"
        and [simp]: "s = C\<langle>u \<cdot> \<sigma>\<rangle>" "t = C\<langle>v \<cdot> \<sigma>\<rangle>" by auto
      from * [THEN bspec, OF cp] show ?thesis by auto
    next
      assume "(t, s) \<in> rstep (CP R)"
      then obtain C \<sigma> u v where cp: "(u, v) \<in> CP R"
        and [simp]: "t = C\<langle>u \<cdot> \<sigma>\<rangle>" "s = C\<langle>v \<cdot> \<sigma>\<rangle>" by auto
      show ?thesis by (rule join_sym) (insert * [THEN bspec, OF cp], auto)
    qed
  qed
  with peak_imp_join_or_CP [OF assms(1)] show ?thesis by blast
qed

lemma CP2_imp_peak:
  fixes R :: "('f, 'v :: infinite) trs"
  assumes "(s, t) \<in> CP2 R S"
  shows "(s, t) \<in> (rstep R)\<inverse> O (rstep S)"
proof -
  from assms obtain r p r' \<sigma>
    where overlap: "overlap R S r p r'"
    and \<sigma>: "\<sigma> = the_mgu (fst r) (fst r' |_ p)"
    and s: "s = replace_at (fst r') p (snd r) \<cdot> \<sigma>"
    and t: "t = snd r' \<cdot> \<sigma>"
    by (auto simp: CP2_def)
  obtain l\<^sub>1 r\<^sub>1 l\<^sub>2 r\<^sub>2
    where [simp]: "r = (l\<^sub>1, r\<^sub>1)" "r' = (l\<^sub>2, r\<^sub>2)" by (cases r, cases r') auto
  from overlap obtain \<pi>\<^sub>1 \<pi>\<^sub>2 and \<tau> :: "('f, 'v) subst"
    where "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> R" and "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> S"
    and p: "p \<in> funposs l\<^sub>2"
    and unif: "l\<^sub>1 \<cdot> \<tau> = l\<^sub>2 |_ p \<cdot> \<tau>"
    by (auto simp: overlap_def)
  then have "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> rstep R" and "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> rstep S"
    by (metis rstep.id surjective_pairing)+
  then have "(l\<^sub>1, r\<^sub>1) \<in> rstep R" and "(l\<^sub>2, r\<^sub>2) \<in> rstep S"
    by (metis perm_rstep_imp_rstep rule_pt.permute_prod.simps)+
  then have "((ctxt_of_pos_term p l\<^sub>2)\<langle>l\<^sub>1\<rangle> \<cdot> \<sigma>, s) \<in> rstep R"
    and "(l\<^sub>2 \<cdot> \<sigma>, t) \<in> rstep S" by (auto simp: s t)
  moreover 
  have "replace_at (l\<^sub>2) p (l\<^sub>1) \<cdot> \<sigma> = l\<^sub>2 \<cdot> \<sigma>"
    using the_mgu [OF unif] and funposs_imp_poss [OF p]
    by (auto simp: \<sigma>) (metis replace_at_ident subst_apply_term_ctxt_apply_distrib)
  ultimately
  show ?thesis by (auto) blast
qed

lemma WCR_imp_CP_join:
  assumes "(rstep R)\<inverse> O (rstep R) \<subseteq> (rstep R)\<^sup>\<down>"
   and "(s, t) \<in> CP R"
  shows "(s, t) \<in> (rstep R)\<^sup>\<down>"
  using CP2_imp_peak [OF assms(2)] and assms(1) by blast

text {*The Critical Pair Lemma*}
lemma CP:
  assumes "\<forall>(l, r)\<in>R. vars_term r \<subseteq> vars_term l"
  shows "(rstep R)\<inverse> O (rstep R) \<subseteq> (rstep R)\<^sup>\<down> \<longleftrightarrow> (\<forall>(s, t)\<in>CP R. (s, t) \<in> (rstep R)\<^sup>\<down>)"
  using CP_join_imp_peak_join [OF assms] and WCR_imp_CP_join [of R] by auto

(*TODO: move*)
lemma WCR_alt_def:
  "WCR (rstep R) \<longleftrightarrow> (rstep R)\<inverse> O (rstep R) \<subseteq> (rstep R)\<^sup>\<down>"
  by (auto simp: WCR_defs simp del: reverseTrs)

text {*A terminating TRS is confluent iff all critical pairs are joinable.*}
corollary
  assumes "\<forall>(l, r)\<in>R. vars_term r \<subseteq> vars_term l"
    and "SN (rstep R)"
  shows "CR (rstep R) \<longleftrightarrow> (\<forall>(s, t)\<in>CP R. (s, t) \<in> (rstep R)\<^sup>\<down>)"
  using Newman [OF assms(2)] and CP [OF assms(1)]
  unfolding WCR_alt_def [symmetric]
  by (simp) (metis CR_on_def WCR_onI r_into_rtrancl)

definition permute_pos :: "'a perm \<Rightarrow> pos \<Rightarrow> pos" where
  "permute_pos \<pi> (p :: pos) = p"

interpretation pos_pure!: pure permute_pos
  by (default) (simp_all add: permute_pos_def)

adhoc_overloading
  PERMUTE permute_pos

interpretation pos_rule_pt!: prod_pt permute_pos rule_pt.permute_prod ..

interpretation overlap_pt!:
  prod_pt rule_pt.permute_prod "prod_pt.permute_prod permute_pos rule_pt.permute_prod" ..

adhoc_overloading
  PERMUTE overlap_pt.permute_prod

lemma permuted_rule_in_variants:
  assumes "(p \<bullet> s, p \<bullet> t) \<in> R"
  shows "(s, t) \<in> rule_pt.variants `` R"
  using assms
  by (auto simp: rule_pt.variants_def)
     (metis rule_pt.permute_minus_cancel(2) rule_pt.permute_prod.simps)

lemma in_funposs_eqvt:
  assumes "p \<in> funposs t"
  shows "p \<in> funposs (\<pi> \<bullet> t)"
  using assms by (induct t arbitrary: p) auto

instance list :: (type) infinite by (default) (rule infinite_UNIV_listI)

lemma
  fixes R :: "('f, string) trs"
  assumes "\<forall>(l, r)\<in>R. vars_term r \<subseteq> vars_term l"
  shows "(rstep R)\<inverse> O (rstep R) \<subseteq> (rstep R)\<^sup>\<down> \<longleftrightarrow> (\<forall>(s, t)\<in>CP R. (s, t) \<in> (rstep R)\<^sup>\<down>)"
  using assms by (rule CP)

text {*Two terms are variants iff they are substitution instances of each other.*}
lemma term_variants_iff:
  fixes s t :: "('f, 'v :: infinite) term"
  shows "(\<exists>\<pi>. \<pi> \<bullet> s = t) \<longleftrightarrow> (\<exists>(\<sigma> :: ('f, 'v) subst) (\<tau> :: ('f, 'v) subst). s \<cdot> \<sigma> = t \<and> t \<cdot> \<tau> = s)"
  (is "?L = ?R")
proof
  assume "?L"
  then obtain \<pi> where *: "\<pi> \<bullet> s = t" ..
  have "s \<cdot> (Var \<circ> Rep_perm \<pi>) = t" by (simp add: *)
  moreover have "t \<cdot> (Var \<circ> Rep_perm (-\<pi>)) = s" by (simp add: * [symmetric])
  ultimately show "?R" by blast
next
  assume "?R"
  then obtain \<sigma> \<tau> :: "('f, 'v) subst"
    where variants: "s \<cdot> \<sigma> = t" "t \<cdot> \<tau> = s" by blast
  from variants_imp_bij_betw_vars [OF variants]
    have "bij_betw (the_Var \<circ> \<sigma>) (vars_term s) (vars_term t)" .
  from bij_betw_extend [OF this, of UNIV] obtain g :: "'v \<Rightarrow> 'v"
    where "finite {x. g x \<noteq> x}" and *: "\<forall>x\<in>vars_term s. g x = (the_Var \<circ> \<sigma>) x" and "bij g" by auto
  then have "g \<in> perms" by (simp add: perms_def)
  then have **: "s \<cdot> (Var \<circ> g) = Abs_perm g \<bullet> s" by (rule term_apply_subst_Var_Abs_perm)
  from * have "\<forall>x\<in>vars_term s. (Var \<circ> g) x = \<sigma> x"
    using Var_the_Var_ident and variants_imp_is_Var [OF variants] by simp
  from term_subst_eq_conv [THEN iffD2, OF this]
    show "?L" unfolding variants and ** ..
qed

text {*The range (in a functional sense) of a substituion.*}
definition subst_fun_range :: "('f, 'v, 'w) gsubst \<Rightarrow> 'w set"
where
  "subst_fun_range \<sigma> = \<Union>(vars_term ` range \<sigma>)"

(*TODO: move*)
lemma subst_variants_imp_is_Var:
  assumes "\<sigma> \<circ>s \<sigma>' = \<tau>" and "\<tau> \<circ>s \<tau>' = \<sigma>"
  shows "\<forall>x\<in>subst_fun_range \<sigma>. is_Var (\<sigma>' x)"
  using assms by (auto simp: subst_compose_def subst_fun_range_def) (metis variants_imp_is_Var)

(*Unification Revisited, Proposition 4.13.*)
lemma
  fixes \<mu> \<theta> :: "('f, 'v :: infinite) subst"
  assumes variants: "\<mu> \<circ>s \<sigma>' = \<theta>" "\<theta> \<circ>s \<tau>' = \<mu>"
    and finite: "finite (subst_domain \<mu>)"
    and idemp: "\<theta> \<circ>s \<theta> = \<theta>"
  shows "\<exists>\<pi>. \<pi> \<bullet> \<theta> = \<mu>"
proof -
  def \<alpha>' \<equiv> "\<lambda>x. if x \<in> subst_domain \<mu> - subst_domain \<theta> then \<mu> x else Var x"

  have [simp]: "\<And>x y. y \<in> vars_term (\<theta> x) \<Longrightarrow> y \<notin> subst_domain \<theta>"
    using idemp by (auto simp: subst_domain_def) (metis subst_apply_term_empty subst_compose_def term_subst_eq_rev)

  have "\<mu> = \<theta> \<circ>s \<tau>'" by (simp add: variants)
  then have "\<theta> \<circ>s \<mu> = \<theta> \<circ>s \<theta> \<circ>s \<tau>'" by (simp add: ac_simps)
  also have "\<dots> = \<theta> \<circ>s \<tau>'" by (simp add: idemp)
  also have "\<dots> = \<mu>" by (simp add: variants)
  finally have "\<mu> = \<theta> \<circ>s \<mu>" ..
  also have "\<dots> = \<theta> \<circ>s \<alpha>'"
  proof (rule ext)
    fix x
    have "\<forall>y \<in> vars_term (\<theta> x). \<mu> y = \<alpha>' y"
    proof
      fix y
      assume "y \<in> vars_term (\<theta> x)"
      then show "\<mu> y = \<alpha>' y" by (simp add: \<alpha>'_def) (simp add: subst_domain_def)
    qed
    then have "\<theta> x \<cdot> \<mu> = \<theta> x \<cdot> \<alpha>'" by (metis term_subst_eq_conv)
    then show "(\<theta> \<circ>s \<mu>) x = (\<theta> \<circ>s \<alpha>') x" by (simp add: subst_compose)
  qed
  finally have "\<mu> = \<theta> \<circ>s \<alpha>'" .
  then have "\<theta> \<circ>s \<alpha>' \<circ>s \<sigma>' = \<theta>" using variants by simp
  have "subst_domain \<theta> \<inter> subst_domain \<alpha>' = {}" by (auto simp: subst_domain_def \<alpha>'_def)
oops

lemma subst_variants_imp_eq:
  assumes "\<sigma> \<circ>s \<sigma>' = \<tau>" and "\<tau> \<circ>s \<tau>' = \<sigma>"
  shows "\<And>x. \<sigma> x \<cdot> \<sigma>' = \<tau> x" "\<And>x. \<tau> x \<cdot> \<tau>' = \<sigma> x"
  using assms by (metis subst_compose)+

lemma subst_variants_id:
  assumes variants: "\<sigma> \<circ>s \<sigma>' = \<tau>" "\<tau> \<circ>s \<tau>' = \<sigma>"
    and x: "x \<in> \<Union>(vars_term ` \<sigma> ` V)" (is "x \<in> ?D")
  shows "(the_Var \<circ> \<tau>') ((the_Var \<circ> \<sigma>') x) = x" (is "?\<tau>' (?\<sigma>' x) = x")
proof -
  note * [simp] = subst_variants_imp_eq [OF variants]
  { fix y
    assume **: "x \<in> vars_term (\<sigma> y)"
    then have var: "is_Var (\<sigma>' x)"
      using variants_imp_is_Var [OF *] by simp
    have "\<And>\<mu>. \<mu> x = Var x \<or> \<sigma> y \<cdot> \<mu> \<noteq> \<sigma> y"
      using ** by (metis subst_apply_term_empty term_subst_eq_rev)
    then have "\<And>\<mu> \<nu>. (\<nu> x) \<cdot> \<mu> = Var x \<or> \<sigma> y \<cdot> \<nu> \<cdot> \<mu> \<noteq> \<sigma> y"
      by (metis subst_apply_term.simps(1) subst_compose_apply_term_distrib)
    then have "?\<tau>' (?\<sigma>' x) = x"
      using var and * by (metis Var_the_Var_ident subst_apply_term.simps(1) the_Var.simps o_def) }
  then show ?thesis using x by auto
qed

lemma subst_variants_image_subset:
  assumes variants: "\<sigma> \<circ>s \<sigma>' = \<tau>" "\<tau> \<circ>s \<tau>' = \<sigma>"
  shows "(the_Var \<circ> \<sigma>') ` \<Union>(vars_term ` \<sigma> ` V) \<subseteq> \<Union>(vars_term ` \<tau> ` V)"
  (is "?\<sigma>' ` ?S \<subseteq> ?T")
proof
  note * = subst_variants_imp_eq [OF variants]
  have subst_fun_range_subset: "?S \<subseteq> subst_fun_range \<sigma>" by (auto simp: subst_fun_range_def)
  note is_Var = subst_variants_imp_is_Var [OF variants]

  fix x\<^sub>0
  assume "x\<^sub>0 \<in> ?\<sigma>' ` ?S"
  then obtain x where **: "?\<sigma>' x = x\<^sub>0" and "x \<in> ?S" by blast
  then obtain x' where "x \<in> vars_term (\<sigma> x')" and "x' \<in> V" by (auto)
  then have "\<sigma> x' \<unrhd> Var x" by (metis supteq_Var)
  then have "\<sigma> x' \<cdot> \<sigma>' \<unrhd> \<sigma>' x" by auto
  then have "\<tau> x' \<unrhd> \<sigma>' x" by (simp add: *)
  moreover
  from is_Var and subst_fun_range_subset and `x \<in> ?S` have "is_Var (\<sigma>' x)" by auto
  ultimately have "?\<sigma>' x \<in> vars_term (\<tau> x')"
    by (metis Var_the_Var_ident comp_def var_subt_imp_subt_in_vars)
  then have "?\<sigma>' x \<in> ?T" using `x' \<in> V` by (auto simp:)
  then show "x\<^sub>0 \<in> ?T" unfolding ** .
qed

lemma subst_variants_imp_image_eq:
  assumes "\<sigma> \<circ>s \<sigma>' = \<tau>" and "\<tau> \<circ>s \<tau>' = \<sigma>"
  shows "(the_Var \<circ> \<sigma>') ` \<Union>(vars_term ` \<sigma> ` V) = \<Union>(vars_term ` \<tau> ` V)"
  (is "?\<sigma>' ` ?S = ?T")
proof
  show "?\<sigma>' ` ?S \<subseteq> ?T"
    using subst_variants_image_subset [OF assms, of V] by (simp)
next
  let ?\<tau>' = "the_Var \<circ> \<tau>'"
  have "?\<sigma>' ` ?\<tau>' ` ?T \<subseteq> ?\<sigma>' ` ?S"
    using subst_variants_image_subset [OF assms(2, 1), of V] by (metis image_mono)
  moreover have "?\<sigma>' ` ?\<tau>' ` ?T = ?T"
    using subst_variants_id [OF assms(2, 1), of _ V]
    by (auto simp: o_def) (metis, metis (full_types) UN_I image_eqI)
  ultimately show "?T \<subseteq> ?\<sigma>' ` ?S" by simp
qed

text {*If two variables are substitution instances of each other, then they only differ
by a variable renaming.*}
lemma subst_variants_imp_perm:
  fixes \<sigma> \<tau> :: "('f, 'v :: infinite) subst"
  assumes variants: "\<sigma> \<circ>s \<sigma>' = \<tau>" "\<tau> \<circ>s \<tau>' = \<sigma>"
    and finite: "finite (subst_domain \<sigma>)" "finite (subst_domain \<tau>)"
  shows "\<exists>\<pi>. \<pi> \<bullet> \<sigma> = \<tau>"
proof -
  def D \<equiv> "subst_domain \<sigma> \<union> subst_domain \<tau>"
  def D\<^sub>\<sigma> \<equiv> "\<Union>(vars_term ` \<sigma> ` D)"
  def D\<^sub>\<tau> \<equiv> "\<Union>(vars_term ` \<tau> ` D)"
  let ?\<sigma>' = "the_Var \<circ> \<sigma>'"
  let ?\<tau>' = "the_Var \<circ> \<tau>'"

  note * = subst_variants_imp_eq [OF variants]
  note is_Var = subst_variants_imp_is_Var [OF variants]

  have finite: "finite D\<^sub>\<sigma>" using finite by (auto simp: D\<^sub>\<sigma>_def D_def)
  have subst_fun_range_subset: "D\<^sub>\<sigma> \<subseteq> subst_fun_range \<sigma>" by (auto simp: D\<^sub>\<sigma>_def subst_fun_range_def)
  have id: "\<And>x. x \<notin> D \<Longrightarrow> \<sigma>' x = Var x" "\<And>x. x \<notin> D \<Longrightarrow> \<tau>' x = Var x"
           "\<And>x. x \<notin> D \<Longrightarrow> \<sigma> x = Var x" "\<And>x. x \<notin> D \<Longrightarrow> \<tau> x = Var x"
    using * by (auto simp: D_def subst_domain_def) (metis subst_apply_term.simps(1))+

  have "?\<sigma>' ` D\<^sub>\<sigma> = D\<^sub>\<tau>"
    using subst_variants_imp_image_eq [OF variants] by (simp add: D\<^sub>\<sigma>_def D\<^sub>\<tau>_def)
  moreover have "inj_on ?\<sigma>' D\<^sub>\<sigma>"
    by (rule inj_on_inverseI [of _ "?\<tau>'"]) (insert subst_variants_id [OF variants], simp add: D\<^sub>\<sigma>_def)
  ultimately have "bij_betw ?\<sigma>' D\<^sub>\<sigma> D\<^sub>\<tau>" by (auto simp: bij_betw_def)
  from bij_betw_extend [OF this, of UNIV] and finite obtain g
    where "finite {x. g x \<noteq> x}" and g_id: "(\<forall>x\<in>UNIV - (D\<^sub>\<sigma> \<union> D\<^sub>\<tau>). g x = x)"
      and g_eq: "\<forall>x\<in>D\<^sub>\<sigma>. g x = (the_Var \<circ> \<sigma>') x" and "bij g" by blast
  then have perm: "g \<in> perms" by (auto simp: perms_def)
  have "Abs_perm g \<bullet> \<sigma> = \<tau>"
  proof (rule ext)
    fix x
    have "Abs_perm g \<bullet> \<sigma> x = \<tau> x"
    proof (cases "x \<in> D")
      assume "x \<in> D"
      then have "vars_term (\<sigma> x) \<subseteq> D\<^sub>\<sigma>" by (auto simp: D\<^sub>\<sigma>_def)
      with g_eq and term_subst_eq [of "\<sigma> x" "Var \<circ> g" "\<sigma>'"]
        have "\<sigma> x \<cdot> (Var \<circ> g) = \<sigma> x \<cdot> \<sigma>'"
        by auto (metis subst_fun_range_subset(1) Var_the_Var_ident is_Var(1) set_rev_mp)
      with perm and * show ?thesis by (simp add: term_apply_subst_Var_Abs_perm)
    next
      have "D\<^sub>\<tau> - D \<subseteq> D\<^sub>\<sigma> - D"
      proof -
        have [simp]: "?\<tau>' ` (D\<^sub>\<tau> - D) = D\<^sub>\<tau> - D"
          using id by auto (metis (full_types) DiffI imageI image_compose the_Var.simps)
        have "?\<tau>' ` (D\<^sub>\<tau> - D) \<subseteq> ?\<tau>' ` D\<^sub>\<tau>" by blast
        also have "... \<subseteq> D\<^sub>\<sigma>"
          using subst_variants_image_subset [OF variants(2, 1), of D] by (simp add: D\<^sub>\<tau>_def D\<^sub>\<sigma>_def)
        finally show ?thesis by auto
      qed
      then have **: "D\<^sub>\<tau> - D\<^sub>\<sigma> \<subseteq> D" by blast
      assume "x \<notin> D"
      moreover then have "Abs_perm g \<bullet> x = x"
        using g_eq and id and g_id and perm and **
        by (cases "x \<in> D\<^sub>\<sigma>") (auto simp: permute_atom_def Abs_perm_inverse)
      ultimately show ?thesis by (simp add: id)
    qed
    then show "(Abs_perm g \<bullet> \<sigma>) x = \<tau> x" by (simp add: permute_subst_def)
  qed
  then show ?thesis ..
qed

definition
  "CP2_rules_pos R R' r p r' =
    {(replace_at (fst (\<pi>\<^sub>2 \<bullet> r')) p (snd (\<pi>\<^sub>1 \<bullet> r)) \<cdot> \<sigma>, snd (\<pi>\<^sub>2 \<bullet> r') \<cdot> \<sigma>) | \<pi>\<^sub>1 \<pi>\<^sub>2 \<sigma>.
    mgu (fst (\<pi>\<^sub>1 \<bullet> r)) (fst (\<pi>\<^sub>2 \<bullet> r') |_ p) = Some \<sigma> \<and>
    overlap R R' (\<pi>\<^sub>1 \<bullet> r) p (\<pi>\<^sub>2 \<bullet> r')}"

(*Auxiliary definition to obtain the result that only finitely many CPs have to be
considered.*)
definition
  CP2' :: "('f, 'v::infinite) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) rule set set"
where
  "CP2' R R' = {CP2_rules_pos R R' r p r' | r p r'. r \<in> R \<and> r' \<in> R' \<and> p \<in> funposs (fst r')}"

lemma finite_funposs:
  "finite (funposs t)"
  by (induct t) auto

lemma finite_CP2':
  assumes "finite R" and "finite R'"
  shows "finite (CP2' R R')"
proof -
  let ?R = "{(r, p, r'). r \<in> R \<and> r' \<in> R' \<and> p \<in> funposs (fst r')}"
  have *: "CP2' R R' = (\<lambda>(r, p, r'). {(replace_at (fst (\<pi>\<^sub>2 \<bullet> r')) p (snd (\<pi>\<^sub>1 \<bullet> r)) \<cdot> \<sigma>, snd (\<pi>\<^sub>2 \<bullet> r') \<cdot> \<sigma>) |
    \<pi>\<^sub>1 \<pi>\<^sub>2 \<sigma>. mgu (fst (\<pi>\<^sub>1 \<bullet> r)) (fst (\<pi>\<^sub>2 \<bullet> r') |_ p) = Some \<sigma> \<and>
    overlap R R' (\<pi>\<^sub>1 \<bullet> r) p (\<pi>\<^sub>2 \<bullet> r')}) ` ?R"
    by (fastforce simp: CP2'_def CP2_rules_pos_def)
  have "{(r, p, r'). r \<in> R \<and> r' \<in> R' \<and> p \<in> funposs (fst r')} \<subseteq> R \<times> \<Union>(funposs ` lhss R') \<times> R'"
    by auto
  moreover have "finite (R \<times> \<Union>(funposs ` lhss R') \<times> R')"
    using assms and finite_funposs by auto
  ultimately have "finite ?R" by (rule finite_subset)
  then show ?thesis unfolding * by (rule finite_imageI)
qed

lemma vars_rule_eqvt [eqvt]:
  "\<pi> \<bullet> vars_rule r = vars_rule (\<pi> \<bullet> r)"
  by (simp add: vars_rule_def) (metis rule_fs.supp_eqvt supp_vars_rule_eq)

lemma funposs_perm_simp [simp]:
  "funposs (\<pi> \<bullet> t) = funposs t"
  by (induct t) auto

lemma poss_perm_simp [simp]:
  "poss (\<pi> \<bullet> t) = poss t"
  by (induct t) auto


interpretation pos_set_pt!: set_pt permute_pos ..

adhoc_overloading
  PERMUTE pos_set_pt.permute_set

interpretation pos_set_pure!: pure pos_set_pt.permute_set
  by default (simp add: permute_pos_def pos_set_pt.permute_set_def)

lemma funposs_eqvt [eqvt]:
  "\<pi> \<bullet> funposs t = funposs (\<pi> \<bullet> t)"
  by simp

lemma poss_eqvt [eqvt]:
  "\<pi> \<bullet> poss t = poss (\<pi> \<bullet> t)"
  by simp

(*TODO: move*)
lemma PCons_poss_Var [simp]:
  "i <# p \<in> poss (Var x) \<longleftrightarrow> False"
  by simp

lemma subt_at_eqvt [eqvt]:
  "p \<in> poss t \<Longrightarrow> \<pi> \<bullet> (t |_ p) = (\<pi> \<bullet> t) |_ p"
  by (induct t p rule: subt_at.induct)
     (auto simp del: subt_at_PCons_distr)

lemma overlap_permute_rules:
  fixes R :: "('f, 'v :: infinite) trs"
  assumes "overlap R R' r p r'"
  shows "overlap R R' (\<pi> \<bullet> r) p (\<pi> \<bullet> r')"
  (is "overlap R R' ?r p ?r'")
proof -
  obtain \<pi>\<^sub>1' \<pi>\<^sub>2' and \<sigma> :: "('f, 'v) subst"
    where rules: "\<pi>\<^sub>1' \<bullet> r \<in> R" "\<pi>\<^sub>2' \<bullet> r' \<in> R'"
    and disj: "vars_rule r \<inter> vars_rule r' = {}"
    and pos: "p \<in> funposs (fst r')"
    and unif: "fst r \<cdot> \<sigma> = fst r' |_ p \<cdot> \<sigma>"
    and triv: "p = \<epsilon> \<Longrightarrow> \<not> (\<exists>\<pi>. \<pi> \<bullet> r = r')"
    using assms by (auto simp: overlap_def)
  have "(\<pi>\<^sub>1' + -\<pi>) \<bullet> ?r \<in> R" and "(\<pi>\<^sub>2' + -\<pi>) \<bullet> ?r' \<in> R'"
    using rules by auto
  then have "\<exists>\<pi>. \<pi> \<bullet> ?r \<in> R" and "\<exists>\<pi>. \<pi> \<bullet> ?r' \<in> R'" by blast+
  moreover have "vars_rule ?r \<inter> vars_rule ?r' = {}"
    unfolding vars_rule_eqvt [symmetric] using disj
    unfolding atom_set_pt.inter_eqvt [symmetric] by (simp)
  moreover have pos': "p \<in> funposs (fst ?r')" using pos by (simp add: rule_pt.fst_eqvt [symmetric])
  moreover have "fst ?r \<cdot> conjugate_subst \<pi> \<sigma> = fst ?r' |_ p \<cdot> conjugate_subst \<pi> \<sigma>"
    using unif and funposs_imp_poss [OF pos'] and funposs_imp_poss [OF pos]
    by (simp add: eqvt [symmetric])
  moreover have "p = \<epsilon> \<Longrightarrow> \<not> (\<exists>\<pi>. \<pi> \<bullet> ?r = ?r')"
    using triv by (auto) (metis left_minus rule_pt.permute_plus rule_pt.permute_zero)
  ultimately show ?thesis
    by (auto simp: overlap_def)
qed

lemma overlap_permuted_rules:
  fixes R :: "('f, 'v :: infinite) trs"
  assumes "overlap R R' (\<pi> \<bullet> r) p (\<pi> \<bullet> r')" (is "overlap R R' ?r p ?r'")
  shows "overlap R R' r p r'"
proof -
  obtain \<pi>\<^sub>1' \<pi>\<^sub>2' and \<sigma> :: "('f, 'v) subst"
    where rules: "\<pi>\<^sub>1' \<bullet> ?r \<in> R" "\<pi>\<^sub>2' \<bullet> ?r' \<in> R'"
    and disj: "vars_rule ?r \<inter> vars_rule ?r' = {}"
    and pos: "p \<in> funposs (fst ?r')"
    and unif: "fst ?r \<cdot> \<sigma> = fst ?r' |_ p \<cdot> \<sigma>"
    and triv: "p = \<epsilon> \<Longrightarrow> \<not> (\<exists>\<pi>. \<pi> \<bullet> ?r = ?r')"
    using assms by (auto simp: overlap_def)
  have "(\<pi>\<^sub>1' + \<pi>) \<bullet> r \<in> R" and "(\<pi>\<^sub>2' + \<pi>) \<bullet> r' \<in> R'"
    using rules by auto
  then have "\<exists>\<pi>. \<pi> \<bullet> r \<in> R" and "\<exists>\<pi>. \<pi> \<bullet> r' \<in> R'" by blast+
  moreover have "vars_rule r \<inter> vars_rule r' = {}"
    using disj unfolding eqvt [symmetric]
    by (metis atom_set_pt.empty_eqvt atom_set_pt.permute_eq_iff)
  moreover have pos': "p \<in> funposs (fst r')" using pos by (simp add: rule_pt.fst_eqvt [symmetric])
  moreover have "fst r \<cdot> ((Var \<circ> Rep_perm \<pi>) \<circ>s \<sigma>) = fst r' |_ p \<cdot> ((Var \<circ> Rep_perm \<pi>) \<circ>s \<sigma>)"
    using unif and funposs_imp_poss [OF pos'] by (simp add: eqvt [symmetric])
  moreover have "p = \<epsilon> \<Longrightarrow> \<not> (\<exists>\<pi>. \<pi> \<bullet> r = r')"
    using triv by (auto) (metis diff_add_cancel rule_pt.permute_plus)
  ultimately show ?thesis
    unfolding overlap_def by blast
qed

lemma overlap_perm_simp [simp]:
  "overlap R R' (\<pi> \<bullet> r) p (\<pi> \<bullet> r') = overlap R R' r p r'"
  by (metis overlap_permute_rules overlap_permuted_rules)

lemma the_mgu_is_imgu:
  fixes \<sigma> :: "('f, 'v) subst"
  assumes "s \<cdot> \<sigma> = t \<cdot> \<sigma>"
  shows "is_imgu (the_mgu s t) {(s, t)}"
proof -
  from assms have "unifiers {(s, t)} \<noteq> {}" by (force simp: unifiers_def)
  then obtain \<tau> where "mgu s t = Some \<tau>"
    and "the_mgu s t = \<tau>"
    using mgu_complete by (auto simp: the_mgu_def)
  with mgu_sound show ?thesis by blast
qed

lemma is_mgu_imp_perm:
  fixes E :: "('f, 'v :: infinite) equations"
  assumes mgu: "is_mgu \<sigma> E" "is_mgu \<tau> E"
    and finite: "finite (subst_domain \<sigma>)" "finite (subst_domain \<tau>)"
  shows "\<exists>\<pi>. \<pi> \<bullet> \<sigma> = \<tau>"
proof -
  obtain \<sigma>' \<tau>' :: "('f, 'v) subst"
    where "\<sigma> \<circ>s \<sigma>' = \<tau>" and "\<tau> \<circ>s \<tau>' = \<sigma>"
    using mgu by (auto simp: is_mgu_def unifiers_def) metis
  from subst_variants_imp_perm [OF this finite]
    show ?thesis .
qed

lemma unifiers_perm_simp [simp]:
  "\<pi> \<bullet> \<sigma> \<in> unifiers E \<longleftrightarrow> \<sigma> \<in> unifiers E"
  by (auto simp: unifiers_def permute_subst_subst_compose)

lemma inv_Rep_perm_simp [simp]:
  fixes x :: "'v :: infinite"
  shows "-\<pi> \<bullet> Rep_perm \<pi> x = x"
  by (simp add: permute_atom_def Rep_perm_uminus)
     (metis Rep_perm_uminus atom_pt.permute_minus_cancel(2) permute_atom_def)

lemma permute_subst_conjugate_subst [simp]:
  "(\<pi> \<bullet> \<sigma>) \<circ>s conjugate_subst \<pi> \<tau> = \<sigma> \<circ>s (\<tau> \<circ>s (Var \<circ> Rep_perm \<pi>))"
  apply (rule ext)
  apply (simp add: subst_conjugate_pt.permute_fun_def)
  apply (simp add: permute_subst_subst_compose ac_simps)
  apply (simp add: subst_compose_def)
done

text {*Every renaming of an mgu is again an mgu.*}
lemma is_mgu_perm:
  fixes E :: "('f, 'v :: infinite) equations"
  assumes "is_mgu \<sigma> E"
  shows "is_mgu (\<pi> \<bullet> \<sigma>) E"
proof (unfold is_mgu_def, intro conjI ballI)
  show "\<pi> \<bullet> \<sigma> \<in> unifiers E" using assms by (simp add: is_mgu_def)
next
  fix \<tau> :: "('f, 'v) subst"
  assume "\<tau> \<in> unifiers E"
  then obtain \<gamma> :: "('f, 'v) subst"
    where "\<tau> = \<sigma> \<circ>s \<gamma>" using assms by (auto simp: is_mgu_def)
  then have "\<tau> \<circ>s (Var \<circ> Rep_perm \<pi>) = (\<pi> \<bullet> \<sigma>) \<circ>s conjugate_subst \<pi> \<gamma>"
    by (simp add: ac_simps)
  then have "\<tau> \<circ>s (Var \<circ> Rep_perm \<pi>) \<circ>s (Var \<circ> Rep_perm (-\<pi>)) =
    (\<pi> \<bullet> \<sigma>) \<circ>s conjugate_subst \<pi> \<gamma> \<circ>s (Var \<circ> Rep_perm (-\<pi>))" by simp
  then have "\<tau> = (\<pi> \<bullet> \<sigma>) \<circ>s conjugate_subst \<pi> \<gamma> \<circ>s (Var \<circ> Rep_perm (-\<pi>))"
    by (simp add: subst_compose_def)
  then have "\<tau> = (\<pi> \<bullet> \<sigma>) \<circ>s (conjugate_subst \<pi> \<gamma> \<circ>s (Var \<circ> Rep_perm (-\<pi>)))"
    unfolding subst_compose_assoc .
  then show "\<exists>\<gamma>::('f, 'v) subst. \<tau> = (\<pi> \<bullet> \<sigma>) \<circ>s \<gamma>" by blast
qed

text {*Turning a permutation into a substitution.*}
abbreviation "sop \<pi> \<equiv> Var \<circ> Rep_perm \<pi>"

lemma Rep_perm_image:
  "Rep_perm \<pi> ` A = \<pi> \<bullet> A"
  by (metis atom_set_pt.permute_set_eq_image image_cong permute_atom_def)

lemma finite_vars_rule:
  "finite (vars_rule r)"
  by (auto simp: vars_rule_def)

(*Can this be proved abstractly, i.e.,
"finite (supp a) ==> ALL x : supp a. p \<bullet> x = p' \<bullet> x ==> p \<bullet> a = p' \<bullet> a"?*)
lemma vars_term_perm_eq:
  assumes "\<forall>x\<in>vars_term t. \<pi> \<bullet> x = \<pi>' \<bullet> x"
  shows "\<pi> \<bullet> t = \<pi>' \<bullet> t"
  using assms by (induct t) simp_all

lemma vars_rule_perm_eq:
  assumes "\<forall>x\<in>vars_rule r. \<pi> \<bullet> x = \<pi>' \<bullet> x"
  shows "\<pi> \<bullet> r = \<pi>' \<bullet> r"
  using assms by (cases r) (auto simp: vars_rule_def vars_term_perm_eq eqvt)

lemma rule_variants_imp_perm:
  assumes disj: "vars_rule (\<pi>\<^sub>1 \<bullet> r) \<inter> vars_rule (\<pi>\<^sub>2 \<bullet> r') = {}" (is "?V \<inter> ?V' = {}")
                "vars_rule (\<pi>\<^sub>3 \<bullet> r) \<inter> vars_rule (\<pi>\<^sub>4 \<bullet> r') = {}" (is "?W \<inter> ?W' = {}")
  shows "\<exists>\<pi>. \<pi> \<bullet> \<pi>\<^sub>3 \<bullet> r = \<pi>\<^sub>1 \<bullet> r \<and> \<pi> \<bullet> \<pi>\<^sub>4 \<bullet> r' = \<pi>\<^sub>2 \<bullet> r'"
proof -
  let ?f = "Rep_perm (\<pi>\<^sub>1 + -\<pi>\<^sub>3)"
  let ?g = "Rep_perm (\<pi>\<^sub>2 + -\<pi>\<^sub>4)"
  have "bij ?f" by (metis bij_Rep_perm)
  then have bij: "bij_betw ?f ?W (?f ` ?W)" by (auto elim: bij_betw_subset)
  have "bij ?g" by (metis bij_Rep_perm)
  then have bij': "bij_betw ?g ?W' (?g ` ?W')" by (auto elim: bij_betw_subset)
  def f \<equiv> "\<lambda>x. if x \<in> vars_rule (\<pi>\<^sub>3 \<bullet> r) then ?f x else x"
  def g \<equiv> "\<lambda>x. if x \<in> vars_rule (\<pi>\<^sub>4 \<bullet> r') then ?g x else x"
  def h \<equiv> "\<lambda>x. if x \<in> vars_rule (\<pi>\<^sub>3 \<bullet> r) then f x else if x \<in> vars_rule (\<pi>\<^sub>4 \<bullet> r') then g x else x"
  have bij: "bij_betw f ?W (f ` ?W)" using bij by (auto simp: f_def bij_betw_def inj_on_def)
  have bij': "bij_betw g ?W' (g ` ?W')" using bij' by (auto simp: g_def bij_betw_def inj_on_def)
  have *: "f ` ?W' = ?W'" "g ` ?W = ?W" using disj(2) by (auto simp: f_def g_def)
  have **: "h ` ?W' = g ` ?W'" "h ` ?W = f ` ?W" using disj(2) by (auto simp: h_def)
  have ***: "f ` ?W = ?V" "g ` ?W' = ?V'" using disj by (auto simp: f_def g_def eqvt Rep_perm_image)
  have "bij_betw h ?W ?V"
    using bij by (auto simp: *** bij_betw_def h_def inj_on_def)
  moreover have "bij_betw h ?W' ?V'"
    using bij' and disj
    apply (auto simp add: * ** *** bij_betw_def h_def inj_on_def)
    apply (metis "***"(2) imageI)
    apply (metis "***"(2) Compl_eq Diff_eq Diff_triv Int_absorb Int_commute)+
  done
  moreover have "h ` ?W \<inter> h ` ?W' = {}"
    unfolding ** *** using disj(1) .
  ultimately have "bij_betw h (?W \<union> ?W') (?V \<union> ?V')"
    using  bij_betw_combine [of h ?W ?V ?W' ?V']
    unfolding ** *** by blast
  from conjI [THEN bij_betw_extend [OF this, of UNIV, simplified], OF finite_vars_rule finite_vars_rule]
    obtain b
    where "finite {x. b x \<noteq> x}"
    and neq: "\<forall>x\<in>UNIV - ((?W \<union> ?W') \<union> (?V \<union> ?V')). b x = x"
    and eq: "\<forall>x\<in>?W \<union> ?W'. b x = h x"
    and "bij b"
    by auto
  then have perm: "b \<in> perms" by (auto simp: perms_def)
  have "Abs_perm b \<bullet> \<pi>\<^sub>3 \<bullet> r = \<pi>\<^sub>1 \<bullet> r"
  proof -
    have "Abs_perm b \<bullet> \<pi>\<^sub>3 \<bullet> r = Abs_perm ?f \<bullet> \<pi>\<^sub>3 \<bullet> r"
    apply (rule vars_rule_perm_eq)
    apply (insert eq perm)
    apply (simp add: h_def f_def g_def Abs_perm_inverse Rep_perm permute_atom_def split: if_splits)
    done
    also have "\<dots> = \<pi>\<^sub>1 \<bullet> r" by (simp add: Rep_perm_inverse)
    finally show ?thesis .
  qed
  moreover have "Abs_perm b \<bullet> \<pi>\<^sub>4 \<bullet> r' = \<pi>\<^sub>2 \<bullet> r'"
  proof -
    have "Abs_perm b \<bullet> \<pi>\<^sub>4 \<bullet> r' = Abs_perm ?g \<bullet> \<pi>\<^sub>4 \<bullet> r'"
    apply (rule vars_rule_perm_eq)
    apply (insert eq perm disj)
    apply (auto simp add: h_def f_def g_def Abs_perm_inverse Rep_perm permute_atom_def split: if_splits)
    done
    also have "\<dots> = \<pi>\<^sub>2 \<bullet> r'" by (simp add: Rep_perm_inverse)
    finally show ?thesis .
  qed  
  ultimately show ?thesis by blast
qed

text {*Overlaps that origin from the same rules at the same position are variants.*}
lemma same_rule_pos_overlap_imp_perm:
  fixes R :: "('f, 'v :: infinite) trs"
  assumes "overlap R R' (\<pi>\<^sub>1 \<bullet> r) p (\<pi>\<^sub>2 \<bullet> r')"
    and "overlap R R' (\<pi>\<^sub>3 \<bullet> r) p (\<pi>\<^sub>4 \<bullet> r')"
  shows "\<exists>\<pi>. \<pi> \<bullet> (\<pi>\<^sub>1 \<bullet> r) = \<pi>\<^sub>3 \<bullet> r \<and> \<pi> \<bullet> (\<pi>\<^sub>2 \<bullet> r') = \<pi>\<^sub>4 \<bullet> r'"
proof -
  from assms obtain \<tau> \<tau>' :: "('f, 'v) subst"
    where "vars_rule (\<pi>\<^sub>1 \<bullet> r) \<inter> vars_rule (\<pi>\<^sub>2 \<bullet> r') = {}"
      and "vars_rule (\<pi>\<^sub>3 \<bullet> r) \<inter> vars_rule (\<pi>\<^sub>4 \<bullet> r') = {}"
    by (auto simp: overlap_def dest!: funposs_imp_poss)
  from rule_variants_imp_perm [OF this(2, 1)] show ?thesis .
qed

lemma poss_perm_prod_simps [simp]:
  "poss (fst (\<pi> \<bullet> r)) = poss (fst r)"
  "poss (snd (\<pi> \<bullet> r)) = poss (snd r)"
  by (case_tac [!] r) (auto simp: eqvt)

lemma ctxt_of_pos_term_eqvt [eqvt]:
  assumes "p \<in> poss t"
  shows "\<pi> \<bullet> (ctxt_of_pos_term p t) = ctxt_of_pos_term p (\<pi> \<bullet> t)"
  using assms by (induct t arbitrary: p) (auto simp: take_map drop_map)

lemma mgu_finite_subst_domain:
  "mgu s t = Some \<sigma> \<Longrightarrow> finite (subst_domain \<sigma>)"
  by (cases "unify [(s, t)] []")
     (auto simp: finite_subst_domain_subst_of)

lemma finite_subst_domain_sop:
  "finite (subst_domain (sop \<pi>))"
  by (auto simp: subst_domain_def finite_Rep_perm)

text {*Critical pairs that origin from overlaps which are variants are also variants
of each other.*}
lemma overlap_variants_imp_CP_variants:
  fixes R R' :: "('f, 'v :: infinite) trs"
  assumes ol: "overlap R R' (\<pi>\<^sub>1 \<bullet> r) p (\<pi>\<^sub>2 \<bullet> r')"
              "overlap R R' (\<pi>\<^sub>3 \<bullet> r) p (\<pi>\<^sub>4 \<bullet> r')"
    and mgu: "mgu (fst (\<pi>\<^sub>1 \<bullet> r)) (fst (\<pi>\<^sub>2 \<bullet> r') |_ p) = Some \<sigma>" (is "mgu ?s ?t = _")
             "mgu (fst (\<pi>\<^sub>3 \<bullet> r)) (fst (\<pi>\<^sub>4 \<bullet> r') |_ p) = Some \<sigma>'" (is "mgu ?u ?v = _")
  shows "\<exists>\<pi>.
    replace_at (fst (\<pi>\<^sub>2 \<bullet> r')) p (snd (\<pi>\<^sub>1 \<bullet> r)) \<cdot> \<sigma> =
    \<pi> \<bullet> (replace_at (fst (\<pi>\<^sub>4 \<bullet> r')) p (snd (\<pi>\<^sub>3 \<bullet> r)) \<cdot> \<sigma>') \<and>
    snd (\<pi>\<^sub>2 \<bullet> r') \<cdot> \<sigma> = \<pi> \<bullet> (snd (\<pi>\<^sub>4 \<bullet> r') \<cdot> \<sigma>')"
proof -
  have "p \<in> poss (fst (\<pi>\<^sub>2 \<bullet> r'))" using ol(1) by (auto simp: overlap_def dest: funposs_imp_poss)
  then have p: "p \<in> poss (fst r')" by (simp add: eqvt [symmetric])
  have mgu': "is_mgu \<sigma> {(?s, ?t)}" "is_mgu \<sigma>' {(?u, ?v)}"
    using mgu_sound [THEN is_imgu_imp_is_mgu] and mgu by auto
  have unif: "?s \<cdot> \<sigma> = ?t \<cdot> \<sigma>" "?u \<cdot> \<sigma>' = ?v \<cdot> \<sigma>'"
    using mgu' by (auto simp: is_mgu_def)
  def s \<equiv> "fst r" and t \<equiv> "fst r' |_ p"
  have unif: "(\<pi>\<^sub>1 \<bullet> s) \<cdot> \<sigma> = (\<pi>\<^sub>2 \<bullet> t) \<cdot> \<sigma>" "(\<pi>\<^sub>3 \<bullet> s) \<cdot> \<sigma>' = (\<pi>\<^sub>4 \<bullet> t) \<cdot> \<sigma>'"
    using unif and p by (simp add: s_def t_def eqvt)+
  have mgu': "is_mgu \<sigma> {(\<pi>\<^sub>1 \<bullet> s, \<pi>\<^sub>2 \<bullet> t)}" "is_mgu \<sigma>' {(\<pi>\<^sub>3 \<bullet> s, \<pi>\<^sub>4 \<bullet> t)}"
    using mgu' and p by (simp add: s_def t_def eqvt)+
  from same_rule_pos_overlap_imp_perm [OF ol(2, 1)] obtain \<pi> :: "'v perm"
    where 0: "\<pi>\<^sub>1 \<bullet> r = \<pi> \<bullet> \<pi>\<^sub>3 \<bullet> r" "\<pi>\<^sub>2 \<bullet> r' = \<pi> \<bullet> \<pi>\<^sub>4 \<bullet> r'" by metis
  then have *: "\<pi>\<^sub>1 \<bullet> s = \<pi> \<bullet> \<pi>\<^sub>3 \<bullet> s" "\<pi>\<^sub>2 \<bullet> t = \<pi> \<bullet> \<pi>\<^sub>4 \<bullet> t"
    using p by (simp_all add: s_def t_def eqvt)
  then have **: "\<pi>\<^sub>3 \<bullet> s = (-\<pi>) \<bullet> \<pi>\<^sub>1 \<bullet> s" "\<pi>\<^sub>4 \<bullet> t = (-\<pi>) \<bullet> \<pi>\<^sub>2 \<bullet> t" by auto

  have "is_mgu (sop \<pi> \<circ>s \<sigma>) {(\<pi>\<^sub>3 \<bullet> s, \<pi>\<^sub>4 \<bullet> t)}"
  proof (unfold is_mgu_def, intro conjI ballI)
    fix \<tau> :: "('f, 'v) subst"
    assume "\<tau> \<in> unifiers {(\<pi>\<^sub>3 \<bullet> s, \<pi>\<^sub>4 \<bullet> t)}"
    then have "\<pi>\<^sub>3 \<bullet> s \<cdot> \<tau> = \<pi>\<^sub>4 \<bullet> t \<cdot> \<tau>" by (simp add: unifiers_def)
    then have "(-\<pi>) \<bullet> \<pi>\<^sub>1 \<bullet> s \<cdot> \<tau> = (-\<pi>) \<bullet> \<pi>\<^sub>2 \<bullet> t \<cdot> \<tau>" by (simp add: ** )
    then have "\<pi>\<^sub>1 \<bullet> s \<cdot> sop (-\<pi>) \<circ>s \<tau> = \<pi>\<^sub>2 \<bullet> t \<cdot> sop (-\<pi>) \<circ>s \<tau>" by simp
    with mgu'(1) obtain \<mu> :: "('f, 'v) subst"
      where "sop (-\<pi>) \<circ>s \<tau> = \<sigma> \<circ>s \<mu>"
      unfolding is_mgu_def and unifiers_def by force
    then have "sop \<pi> \<circ>s sop (-\<pi>) \<circ>s \<tau> = sop \<pi> \<circ>s \<sigma> \<circ>s \<mu>" by (simp add: ac_simps)
    then have "\<tau> = sop \<pi> \<circ>s \<sigma> \<circ>s \<mu>" by (simp add: subst_compose_def)
    then show "\<exists>\<mu>. \<tau> = sop \<pi> \<circ>s \<sigma> \<circ>s \<mu>" by blast
  qed (simp add: unif ** unifiers_def)
  moreover have "finite (subst_domain (sop \<pi> \<circ>s \<sigma>))"
  proof -
    have "subst_domain (sop \<pi> \<circ>s \<sigma>) \<subseteq> subst_domain (sop \<pi> :: ('f, 'v) subst) \<union> subst_domain \<sigma>"
      using subst_domain_compose [of "sop \<pi>" \<sigma>] using [[show_sorts]] .
    moreover have "finite (subst_domain (sop \<pi>))" by (metis finite_subst_domain_sop)
    moreover have "finite (subst_domain \<sigma>)" using mgu_finite_subst_domain [OF mgu(1)] .
    ultimately show ?thesis by (metis infinite_Un infinite_super)
  qed
  moreover have "finite (subst_domain \<sigma>')" using mgu_finite_subst_domain [OF mgu(2)] .
  ultimately obtain \<pi>' where 1: "\<pi>' \<bullet> \<sigma>' = sop \<pi> \<circ>s \<sigma>"
    using is_mgu_imp_perm [OF mgu'(2)] by blast

  have "replace_at (fst (\<pi>\<^sub>2 \<bullet> r') \<cdot> \<sigma>) p (snd (\<pi>\<^sub>1 \<bullet> r) \<cdot> \<sigma>) =
    \<pi>' \<bullet> (replace_at (fst (\<pi>\<^sub>4 \<bullet> r')) p (snd (\<pi>\<^sub>3 \<bullet> r)) \<cdot> \<sigma>')"
  proof -
    have "replace_at (fst (\<pi>\<^sub>2 \<bullet> r') \<cdot> \<sigma>) p (snd (\<pi>\<^sub>1 \<bullet> r) \<cdot> \<sigma>) =
      \<pi> \<bullet> replace_at (fst (\<pi>\<^sub>4 \<bullet> r')) p (snd (\<pi>\<^sub>3 \<bullet> r)) \<cdot> \<sigma>"
      using p by (simp add: 0 eqvt ctxt_of_pos_term_subst)
    also have "\<dots> = \<pi>' \<bullet> (replace_at (fst (\<pi>\<^sub>4 \<bullet> r')) p (snd (\<pi>\<^sub>3 \<bullet> r)) \<cdot> \<sigma>')"
      using 1 apply auto
      by (metis permute_subst_subst_compose subst_apply_term_ctxt_apply_distrib subst_compose_apply_term_distrib term_apply_subst_Var_Rep_perm)
    finally show ?thesis .
  qed
  moreover
  have "snd (\<pi>\<^sub>2 \<bullet> r') \<cdot> \<sigma> = \<pi>' \<bullet> (snd (\<pi>\<^sub>4 \<bullet> r') \<cdot> \<sigma>')"
  proof -
    have "snd (\<pi>\<^sub>2 \<bullet> r') \<cdot> \<sigma> = \<pi> \<bullet> snd (\<pi>\<^sub>4 \<bullet> r') \<cdot> \<sigma>" by (simp add: 0 eqvt)
    also have "\<dots> = \<pi>' \<bullet> (snd (\<pi>\<^sub>4 \<bullet> r') \<cdot> \<sigma>')"
      using 1 by (metis permute_subst_subst_compose subst_compose_apply_term_distrib term_apply_subst_Var_Rep_perm)
    finally show ?thesis .
  qed
  ultimately show ?thesis
    by (metis ctxt_of_pos_term_subst p poss_perm_prod_simps(1) subst_apply_term_ctxt_apply_distrib)
qed

lemma funposs_perm_iff [simp]:
  "p \<in> funposs (\<pi> \<bullet> t) \<longleftrightarrow> p \<in> funposs t"
  by (induct t) auto

lemma CP2_Union_CP2':
  "CP2 R R' = \<Union>(CP2' R R')"
proof
  show "\<Union>(CP2' R R') \<subseteq> CP2 R R'"
  proof
    fix s t
    assume "(s, t) \<in> \<Union>(CP2' R R')"
    then obtain r p r' \<pi>\<^sub>1 \<pi>\<^sub>2 \<sigma>
      where "r \<in> R" and "r' \<in> R'" and p: "p \<in> funposs (fst (\<pi>\<^sub>2 \<bullet> r'))"
      and "mgu (\<pi>\<^sub>1 \<bullet> fst r) (\<pi>\<^sub>2 \<bullet> fst r' |_ p) = Some \<sigma>"
      and ol: "overlap R R' (\<pi>\<^sub>1 \<bullet> r) p (\<pi>\<^sub>2 \<bullet> r')"
      and s: "s = replace_at (fst (\<pi>\<^sub>2 \<bullet> r')) p (snd (\<pi>\<^sub>1 \<bullet> r)) \<cdot> \<sigma>"
      and t: "t = snd (\<pi>\<^sub>2 \<bullet> r') \<cdot> \<sigma>"
      by (auto simp: CP2'_def CP2_rules_pos_def eqvt)
    then have "\<sigma> = the_mgu (\<pi>\<^sub>1 \<bullet> fst r) (\<pi>\<^sub>2 \<bullet> fst r' |_ p)" by (simp add: the_mgu_def)
    with ol show "(s, t) \<in> CP2 R R'"
      using funposs_imp_poss [OF p]
      unfolding CP2_def s t by (force simp add: eqvt ctxt_of_pos_term_subst)
  qed
next
  show "CP2 R R' \<subseteq> \<Union>(CP2' R R')"
  proof
    fix s t
    assume "(s, t) \<in> CP2 R R'"
    then obtain r p r' \<sigma>
      where ol: "overlap R R' r p r'"
      and "\<sigma> = the_mgu (fst r) (fst r' |_ p)"
      and s: "s = replace_at (fst r') p (snd r) \<cdot> \<sigma>"
      and t: "t = snd r' \<cdot> \<sigma>"
      by (auto simp: CP2_def)
    then have mgu: "mgu (fst r) (fst r' |_ p) = Some \<sigma>"
      unfolding overlap_def the_mgu_def
      using unify_complete and unify_sound by (force split: option.splits simp: is_imgu_def unifiers_def)
    from ol obtain \<pi>\<^sub>1 \<pi>\<^sub>2
      where "\<pi>\<^sub>1 \<bullet> r \<in> R" (is "?r \<in> R")
      and "\<pi>\<^sub>2 \<bullet> r' \<in> R'" (is "?r' \<in> R'")
      and p: "p \<in> funposs (fst r')" by (auto simp: overlap_def)
    moreover
    have "overlap R R' (-\<pi>\<^sub>1 \<bullet> ?r) p (-\<pi>\<^sub>2 \<bullet> ?r')" using ol by simp
    moreover
    have "s = (ctxt_of_pos_term p (fst (-\<pi>\<^sub>2 \<bullet> ?r')))\<langle>snd (-\<pi>\<^sub>1 \<bullet> ?r)\<rangle> \<cdot> \<sigma>"
      and "t = snd (-\<pi>\<^sub>2 \<bullet> ?r') \<cdot> \<sigma>"
      using funposs_imp_poss [OF p] by (simp_all add: s t ctxt_of_pos_term_subst)
    moreover
    have "mgu (fst (-\<pi>\<^sub>1 \<bullet> ?r)) (fst (-\<pi>\<^sub>2 \<bullet> ?r') |_ p) = Some \<sigma>" using mgu by simp
    moreover have "p \<in> funposs (fst ?r')" using p by (simp add: eqvt [symmetric])
    ultimately show "(s, t) \<in> \<Union>(CP2' R R')"
      unfolding CP2'_def CP2_rules_pos_def by blast
  qed
qed

lemma CP2_rules_pos_perm:
  fixes R R' :: "('f, 'v :: infinite) trs"
  assumes "x \<in> CP2_rules_pos R R' r p r'"
    and "y \<in> CP2_rules_pos R R' r p r'"
  shows "\<exists>\<pi>. \<pi> \<bullet> x = y"
proof -
  from assms [unfolded CP2_rules_pos_def] obtain \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 and \<sigma> \<sigma>' :: "('f, 'v) subst"
  where mgu: "mgu (fst (\<pi>\<^sub>1 \<bullet> r)) (fst (\<pi>\<^sub>2 \<bullet> r') |_ p) = Some \<sigma>"
             "mgu (fst (\<pi>\<^sub>3 \<bullet> r)) (fst (\<pi>\<^sub>4 \<bullet> r') |_ p) = Some \<sigma>'"
    and ol: "overlap R R' (\<pi>\<^sub>1 \<bullet> r) p (\<pi>\<^sub>2 \<bullet> r')"
            "overlap R R' (\<pi>\<^sub>3 \<bullet> r) p (\<pi>\<^sub>4 \<bullet> r')"
    and x: "x = (replace_at (fst (\<pi>\<^sub>2 \<bullet> r')) p (snd (\<pi>\<^sub>1 \<bullet> r)) \<cdot> \<sigma>, snd (\<pi>\<^sub>2 \<bullet> r') \<cdot> \<sigma>)"
    and y: "y = (replace_at (fst (\<pi>\<^sub>4 \<bullet> r')) p (snd (\<pi>\<^sub>3 \<bullet> r)) \<cdot> \<sigma>', snd (\<pi>\<^sub>4 \<bullet> r') \<cdot> \<sigma>')"
    by blast
  show ?thesis
    using overlap_variants_imp_CP_variants [OF ol(2, 1) mgu(2, 1)]
    by (auto simp: x y eqvt)
qed

lemma perm_rstep_conv [simp]:
  "\<pi> \<bullet> p \<in> rstep R \<longleftrightarrow> p \<in> rstep R"
  by (metis rstep_eqvt rstep_permute trs_pt.mem_permute_iff)

lemma perm_rstep_perm:
  assumes "(\<pi> \<bullet> s, t) \<in> rstep R"
  shows "\<exists>u. t = \<pi> \<bullet> u"
  using assms by (metis term_pt.permute_minus_cancel(1))

lemma perm_rsteps_perm:
  assumes "(\<pi> \<bullet> s, t) \<in> (rstep R)\<^sup>*"
  shows "\<exists>u. t = \<pi> \<bullet> u"
  using assms by (metis term_pt.permute_minus_cancel(1))

lemma perm_rsteps_conv [simp]:
  "\<pi> \<bullet> p \<in> (rstep R)\<^sup>* \<longleftrightarrow> p \<in> (rstep R)\<^sup>*"
  by (metis perm_rstep_conv rstep_rtrancl_idemp)

lemma perm_join_conv [simp]:
  "\<pi> \<bullet> p \<in> (rstep R)\<^sup>\<down> \<longleftrightarrow> p \<in> (rstep R)\<^sup>\<down>" (is "?L = ?R")
proof
  assume "?L"
  then obtain u
    where "(fst (\<pi> \<bullet> p), u) \<in> (rstep R)\<^sup>*" and "(snd (\<pi> \<bullet> p), u) \<in> (rstep R)\<^sup>*"
    by (metis joinE surjective_pairing)
  moreover then obtain v
    where [simp]: "u = \<pi> \<bullet> v" by (auto dest: perm_rsteps_perm simp: eqvt [symmetric])
  ultimately have "(fst p, v) \<in> (rstep R)\<^sup>*" and "(snd p, v) \<in> (rstep R)\<^sup>*"
    by (simp_all add: eqvt [symmetric])
  then show "?R" by (metis joinI surjective_pairing)
next
  assume "?R"
  then obtain u
    where "(fst p, u) \<in> (rstep R)\<^sup>*" and "(snd p, u) \<in> (rstep R)\<^sup>*"
    by (metis joinE surjective_pairing)
  then have "\<pi> \<bullet> (fst p, u) \<in> (rstep R)\<^sup>*" and "\<pi> \<bullet> (snd p, u) \<in> (rstep R)\<^sup>*" by auto
  then show "?L" by (metis joinI rule_pt.permute_prod_eqvt surjective_pairing)
qed

lemma CP2_rules_pos_join:
  assumes "x \<in> CP2_rules_pos R R' r p r'"
    and "y \<in> CP2_rules_pos R R' r p r'"
  shows "x \<in> (rstep S)\<^sup>\<down> \<longleftrightarrow> y \<in> (rstep S)\<^sup>\<down>"
  using CP2_rules_pos_perm [OF assms] by auto

abbreviation "CP' R \<equiv> CP2' R R"

text {*If @{term R} and @{term R'} are finite, then @{term "CP2' R R'"} is finite
by @{thm finite_CP2'}. Thus it suffices to prove joinability of finitely many
critical pairs in order to conclude joinability of all critical pairs @{term "CP2 R R'"}.*}
lemma CP2'_representatives_join_imp_CP2_join:
  assumes "\<forall>C\<in>CP2' R R'. \<exists>(s, t)\<in>C. (s, t) \<in> (rstep S)\<^sup>\<down>"
  shows "\<forall>(s, t)\<in>CP2 R R'. (s, t) \<in> (rstep S)\<^sup>\<down>"
  using assms
  by (auto simp: CP2'_def CP2_Union_CP2')
     (blast dest: CP2_rules_pos_join [of _ R R' _ _ _ _ S])

end

