(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Innermost_Loops_Impl
imports
  Innermost_Loops
  Trs_Impl
  Q_Restricted_Rewriting
  "$AFP/Transitive-Closure/Transitive_Closure_List_Impl"
begin

context fixed_subst_incr
begin
lemma ident_solve'_code: "ident_solve' cps st = (let cp = conflicts st in
    if (\<exists> (u,v,n) \<in> set cp. is_Fun u) then None
    else if (\<exists>(u,v,n) \<in> set cp. (u,v) \<in> (\<lambda> (u,v,m). (u,v)) ` cps) then None
    else do {
      is \<leftarrow> mapM (\<lambda>(u, v, m). ident_solve' (insert (u, v, m) cps) (u \<cdot> \<mu>, v, Suc m)) (conflicts st);
      Some (max_list (map Suc is))
    })" 
  unfolding ident_solve'.simps[of cps st] Let_def
  by (rule if_cong, force, force, rule if_cong, force, force, force)
end

declare fixed_subst.conflicts.simps[code]
declare fixed_subst_incr.ident_solve'_code[code]
declare fixed_subst_incr.ident_solve_def[code]

definition ident_decision :: "('f,'v)subst_incr \<Rightarrow> ('f,'v)ident_prob \<Rightarrow> bool" where
  "ident_decision \<sigma> ip \<equiv> fixed_subst_incr.ident_solve \<sigma> ip \<noteq> None"

lemma ident_decision[simp]: "ident_decision \<mu> ip = (\<exists> i. fixed_subst.ident_solution (si_subst \<mu>) ip i)"
proof -
  interpret fixed_subst_incr \<mu> .
  show ?thesis unfolding ident_decision_def using ident_solve[of ip] by auto
qed

definition gmatch_decision :: "('f,'v)subst_incr \<Rightarrow> ('f,'v)gmatch_prob \<Rightarrow> bool" where
  "gmatch_decision \<sigma> mp \<equiv> do {
    (smp,i) \<leftarrow> fixed_subst_incr.simplify_mp \<sigma> mp [];
    guard (Ball (set (ident_prob_of_smp smp)) (ident_decision \<sigma>))
  } \<noteq> None"

lemma gmatch_decision[simp]: "gmatch_decision \<mu> mp = (\<exists> sol. fixed_subst.gmatch_solution (si_subst \<mu>) mp sol)" (is "?l = ?r")
proof -
  interpret fixed_subst_incr \<mu> .
  note simp = gmatch_decision_def[of \<mu> mp]
  show ?thesis
  proof (cases "simplify_mp mp []")
    case None
    with simplify_mp_complete[of mp] show ?thesis unfolding simp by auto
  next
    case (Some sol)
    then obtain smp i where sol: "sol = (smp, i)" by force
    note Some = Some[unfolded sol]
    note simp = simp[unfolded Some]
    note simp_sound = simplify_mp_sound[OF Some]
    from simplify_mp_solved_form[OF Some] have solved: "gmatch_solved_form smp"
      unfolding gmatch_solved_form_def by auto
    let ?ips = "ident_prob_of_smp smp"
    show ?thesis
    proof (cases "Ball (set ?ips) (ident_decision \<mu>)")
      case True
      with simp have l: "?l = True" by simp
      {
        fix ip
        assume "ip \<in> set ?ips"
        with True have "ident_decision \<mu> ip" by auto
        hence "\<exists> i. ident_solution ip i" by simp
      }
      hence "\<forall> ip. \<exists> i. ip \<in> set ?ips \<longrightarrow> ident_solution ip i" by auto
      from choice[OF this] obtain "is" where sol: "\<And> ip. ip \<in> set ?ips \<Longrightarrow> ident_solution ip (is ip)" by blast
      def i \<equiv> "max_list (map is ?ips)"
      have "\<exists>\<sigma>. gmatch_solution smp (i, \<sigma>)"
        by (rule ident_prob_of_smp_sound[OF solved], rule exI, rule conjI[OF _ sol], unfold i_def, rule max_list, auto)
      then obtain \<sigma> where "gmatch_solution smp (i, \<sigma>)" ..
      from simp_sound[OF this] l show ?thesis by auto
    next
      case False
      with simp have l: "?l = False" by simp
      from False obtain ip where ip: "ip \<in> set ?ips" and False: "\<not> ident_decision \<mu> ip" by auto
      {
        fix n \<sigma>
        assume "gmatch_solution mp (n,\<sigma>)"
        from simplify_mp_complete[OF this] Some obtain k where "gmatch_solution smp (k, \<sigma>)" by auto
        from ident_prob_of_smp_complete[OF this, of "fst ip" "snd ip"] ip False
        have False by auto
      }
      with l show ?thesis by auto
    qed
  qed
qed

definition match_decision :: "('f,'v)subst_incr \<Rightarrow> ('f,'v)match_prob \<Rightarrow> bool" where
  "match_decision \<mu> mp \<equiv> gmatch_decision \<mu> [mp]"

lemma match_decision[simp]: "match_decision \<mu> mp = (\<exists> sol. fixed_subst.match_solution (si_subst \<mu>) mp sol)"
proof -
  interpret fixed_subst_incr \<mu> .
  show ?thesis unfolding match_decision_def gmatch_decision 
    using gmatch_solution_def by auto
qed

context fixed_subst
begin

lemma W_via_rtrancl: 
  "W t = {(x, y). x \<noteq> y \<and> y \<in> vars_term (\<mu> x)}^* `` vars_term t"
proof -
  obtain V where V: "V = vars_term t" by auto
  let ?M = "\<lambda> i V. \<Union>(vars_term ` (\<mu> ^^ i) ` V)"
  let ?R = "{(x,y) | x y. y \<in> vars_term (\<mu> x)}"
  have "?R^* `` V = (\<Union>n. (?R ^^ n) `` V)"
    unfolding rtrancl_is_UN_relpow by auto
  also have "... = (\<Union> n. ?M n V)"
  proof -
    {
      fix n 
      have "?M n V = (?R ^^ n) `` V"
      proof (induct n)
        case 0
        show ?case by auto
      next
        case (Suc n)
        have "(?R ^^ Suc n) `` V = (?R ^^ n O ?R) `` V" by auto
        also have "... = ?R `` ((?R ^^ n) `` V)" by auto
        also have "... = ?R `` ?M n V" unfolding Suc ..
        also have "... = ?M (Suc n) V" 
        proof -
          have "?M (Suc n) V = vars_term (t \<cdot> \<mu> ^^ Suc n)" 
            unfolding V vars_term_subst ..
          also have "... = vars_term (t \<cdot> \<mu> ^^ n \<cdot> \<mu>)" 
            unfolding subst_power_Suc by auto
          also have "... = ?R `` ?M n V" 
            unfolding V vars_term_subst by auto
          finally show ?thesis by simp
        qed
        finally show ?case by simp
      qed
    } note main = this
    show ?thesis unfolding main ..
  qed
  also have "... = W t" unfolding vars_iteration_def vars_term_subst V by auto
  finally have main: "W t = ?R^* `` V" by simp
  show ?thesis unfolding main rtrancl_r_diff_Id[of ?R, symmetric] unfolding V[symmetric] 
    by (rule arg_cong[where f = "\<lambda> x. x^* `` V"], auto)
qed

lemma V_incr_via_rtrancl:
  "V_incr = {(y, x) | x y. x \<noteq> y \<and> \<mu> x = Var y}^* `` {x. is_Fun (\<mu> x)}"
  (is "_ = ?r^* `` ?s")
proof -
  {
    fix x y n
    assume "(x,y) \<in> ?r^*" 
    then obtain n where "(x,y) \<in> ?r^^n" unfolding rtrancl_is_UN_relpow ..
    hence "(\<mu> ^^ n) y = Var x"
    proof (induct n arbitrary: y)
      case 0 thus ?case by force
    next
      case (Suc n y)
      from Suc(2) obtain z where xz: "(x,z) \<in> ?r ^^ n" and zy: "(z,y) \<in> ?r" by auto
      from zy have zy: "\<mu> y = Var z" by auto
      from Suc(1)[OF xz] have x: "Var x = (\<mu> ^^ n) z" by simp
      show ?case unfolding x substpow.simps
        subst_compose_def zy by simp
    qed
    hence "\<exists> n. (\<mu> ^^ n) y = Var x" by auto
  } note one_dir = this
  {
    fix x
    assume "x \<in> ?r^* `` ?s"
    then obtain y where yx: "(y,x) \<in> ?r^*" and y: "y \<in> ?s" by auto
    from one_dir[OF yx] obtain n where x: "(\<mu> ^^ n) x = Var y" ..
    have "x \<in> V_incr" unfolding v_incr_def
      by (rule, rule exI[of _ "Suc n"], 
        unfold subst_power_Suc, insert y, simp add: x)
  }
  moreover 
  {
    fix x
    let ?N = "\<lambda> x n. is_Fun ((\<mu> ^^ n) x)"
    let ?P = "?N x"
    let ?Q = "\<lambda> x. \<exists> n. ?N x n"
    assume "x \<in> V_incr"
    then obtain n where nvar: "?P n" unfolding v_incr_def by auto
    let ?n = "LEAST n. ?P n"
    have nle: "?P ?n" by (rule LeastI, rule nvar)
    from Least_Suc[of ?P, OF nvar] have "?n > 0" by auto
    then obtain n where n: "?n = Suc n" by (cases ?n, auto)
    hence "n < ?n" by auto
    from not_less_Least[OF this] obtain y where y: "(\<mu> ^^ n) x = Var y"
      by (cases "(\<mu> ^^ n) x", auto)
    have "(\<mu> ^^ ?n) x = \<mu> y" unfolding n subst_power_Suc subst_compose_def y
      by simp
    with nle have ys: "y \<in> ?s" by simp
    have "?Q y" by (rule exI[of _ "Suc 0"], insert ys, auto)
    from y have "(y,x) \<in> ?r^^n" 
    proof (induct n arbitrary: x)
      case (0 x)
      thus ?case by auto
    next
      case (Suc n x)
      note id = Suc(2)[unfolded substpow.simps subst_compose_def]
      from id obtain z where xz: "\<mu> x = Var z" and zy: "(\<mu> ^^ n) z = Var y"
        by (cases "\<mu> x", auto)
      from Suc(1)[OF zy] have "(y,z) \<in> ?r ^^ n" .
      moreover have "(z,x) \<in> ?r"
      proof(rule, intro exI, rule conjI[OF refl], rule conjI[OF _ xz])
        show "x \<noteq> z"
        proof
          assume "x = z"
          with xz id have x: "\<mu> x = Var x" and xy: "(\<mu> ^^ n) x = Var y" by auto
          {
            fix n
            have "(\<mu> ^^ n) x = Var x" 
              by (induct n, insert x, auto simp: subst_compose_def)
          } note x = this
          have "\<mu> y = Var y \<cdot> \<mu>" by simp
          also have "... = (\<mu> ^^ (Suc n)) x" unfolding xy[symmetric]
            subst_power_Suc subst_compose_def ..
          also have "... = Var x" unfolding x ..
          finally show False using ys by auto
        qed
      qed
      ultimately show ?case by auto
    qed
    hence "(y,x) \<in> ?r^*" unfolding rtrancl_is_UN_relpow by auto
    with ys have "x \<in> ?r^* `` ?s" by auto
  }
  ultimately show ?thesis by blast
qed
end

definition v_incr_impl where
  "v_incr_impl d \<equiv>
    let 
      (ixs, uxs) = partition (\<lambda>(x, t). is_Fun t) d;
      rel = map (\<lambda>(x, y). (the_Var y, x)) uxs;
      start = map fst ixs
    in rtrancl_list_impl rel start"

lemma v_incr_impl: "set (v_incr_impl (mk_subst_domain \<mu>)) = v_incr (mk_subst Var \<mu>)"
proof -
  let ?V = "v_incr_impl (mk_subst_domain \<mu>)"
  let ?\<mu> = "mk_subst Var \<mu>"
  let ?R = "(\<lambda>(x, y). (the_Var y, x)) `
              {x \<in> (\<lambda>x. (x, mk_subst Var \<mu> x)) ` subst_domain (mk_subst Var \<mu>).
               \<not> (case x of (x, t) \<Rightarrow> is_Fun t)}"
  let ?r = "{(y,x) | x y. x \<noteq> y \<and> ?\<mu> x = Var y}"
  let ?S = "fst ` {x \<in> (\<lambda>x. (x, mk_subst Var \<mu> x)) ` subst_domain (mk_subst Var \<mu>).
            case x of (x, t) \<Rightarrow> is_Fun t}"
  let ?s = "{ x. is_Fun (?\<mu> x)}"
  have idR: "?R = ?r" unfolding subst_domain_def by force
  have idS: "?S = ?s" unfolding subst_domain_def by force
  have "set ?V = ?r^* `` ?s"
    unfolding v_incr_impl_def Let_def
    unfolding partition_filter_conv split o_def 
    unfolding rtrancl_list_impl set_map set_filter mk_subst_domain
    unfolding split
    unfolding idS idR by auto
  also have "... = v_incr ?\<mu>" unfolding fixed_subst.V_incr_via_rtrancl ..
  finally show ?thesis .
qed

definition W_impl :: "('f, 'v) substL \<Rightarrow> ('f, 'v) term \<Rightarrow> 'v list" where
  "W_impl d \<equiv>
    let 
      filt = filter (\<lambda>(x,y). x \<noteq> y);
      xvs = concat (map (\<lambda> (x,t). (map (\<lambda> y. (x,y)) (vars_term_list t))) d);
      rel = filt xvs;
      rtran = rtrancl_list_impl rel
    in (\<lambda>t. rtran (vars_term_list t))"

lemma W_impl: "set (W_impl (mk_subst_domain \<mu>) t) = vars_iteration (mk_subst Var \<mu>) t" 
proof -  
  interpret fixed_subst "mk_subst Var \<mu>" .
  let ?I = "set [(x,y)\<leftarrow>concat (map (\<lambda>(x, t). map (Pair x) (vars_term_list t))
        (mk_subst_domain \<mu>)) . x \<noteq> y]"
  let ?S = " {(x, y) |x y. x \<noteq> y \<and> y \<in> vars_term (mk_subst Var \<mu> x)}"
  have id: "?I = ?S" 
  proof -
    have "?I \<subseteq> ?S" unfolding set_concat set_map set_filter mk_subst_domain
      by auto
    moreover {
      fix xy
      assume S: "xy \<in> ?S"
      obtain x y where xy: "xy = (x,y)" by (cases xy, auto)
      with S have neq: "x \<noteq> y" and y: "y \<in> vars_term (mk_subst Var \<mu> x)" by auto
      hence "mk_subst Var \<mu> x \<noteq> Var x" by (cases "mk_subst Var \<mu> x", auto)
      hence x: "x \<in> subst_domain (mk_subst Var \<mu>)" unfolding subst_domain_def by auto
      have "(x,y) \<in> ?I"
        unfolding set_filter set_concat set_map split mk_subst_domain
        by (rule, rule, rule, rule, rule, rule refl, rule x, simp_all add: y neq)
      hence "xy \<in> ?I" unfolding xy .
    }
    ultimately show "?I = ?S" by blast
  qed
  show ?thesis
    unfolding W_via_rtrancl
    unfolding W_impl_def Let_def
    unfolding rtrancl_list_impl
    unfolding id vars_term_list_sound 
  by auto
qed


lift_definition subst_incr :: "('f,'v)substL \<Rightarrow> ('f,'v)subst_incr" 
  is "\<lambda> \<sigma>. let dom = mk_subst_domain \<sigma> in (mk_subst Var \<sigma>, set (v_incr_impl dom), W_impl dom)" 
  unfolding Let_def v_incr_impl 
  using finite_mk_subst W_impl by force

lemma si_subst_subst_incr: "si_subst (subst_incr \<sigma>) = mk_subst Var \<sigma>"
  by (transfer, simp add: Let_def)

fun match_prob_of_rp_impl :: "('f,'v)subst_incr \<Rightarrow> ('f,'v)redex_prob \<Rightarrow> ('f,'v)match_prob list" where
  "match_prob_of_rp_impl \<mu> (t, Var x) = [(t, Var x)]"
| "match_prob_of_rp_impl \<mu> (t, l) = (
    let
      sterms = remdups (t # map (si_subst \<mu>) (si_W \<mu> t));
      uterms = concat (map (filter is_Fun \<circ> supteq_list) sterms)
    in map (\<lambda>u. (u, l)) (remdups uterms))"

lemma (in fixed_subst_incr) match_prob_of_rp_impl:
  "set (match_prob_of_rp_impl \<mu>_incr (t,l)) =
    match_prob_of_rp (t,l)"
proof (cases l)
  case (Var x)
  show ?thesis unfolding match_prob_of_rp_def Var by simp
next
  case (Fun f ls)
  show ?thesis unfolding Fun match_prob_of_rp_def match_prob_of_rp_impl.simps
    unfolding Let_def split term.simps unfolding si_W[symmetric]
    by auto 
qed

definition redex_decision where
  "redex_decision \<mu> rp \<equiv>
      (\<exists> mp \<in> set (match_prob_of_rp_impl \<mu> rp). match_decision \<mu> mp)"

lemma redex_decision: "redex_decision \<mu> rp = (\<exists> sol. fixed_subst.redex_solution (si_subst \<mu>) rp sol)" (is "?l = ?r")
proof -
  interpret fixed_subst_incr \<mu> .
  let ?\<mu> = "si_subst \<mu>"
  obtain t l where rp: "rp = (t,l)" by force
  show ?thesis
  proof
    assume ?r
    then obtain i \<sigma> C where "redex_solution (t,l) (i,\<sigma>,C)" unfolding rp by auto
    from match_prob_of_rp_complete[OF this]
    obtain \<tau> u j
      where mem: "(u,l) \<in> match_prob_of_rp (t,l)"
      and sol: "match_solution (u,l) (j,\<tau>)" by auto
    show ?l unfolding rp redex_decision_def Let_def 
      set_map match_prob_of_rp_impl
      using match_decision[of \<mu> "(u,l)"] sol mem by auto
  next
    assume ?l
    from this[unfolded rp redex_decision_def Let_def set_map match_prob_of_rp_impl]
    obtain m p where mp: "(m,p) \<in> match_prob_of_rp (t,l)"
      and sol: "match_decision \<mu> (m,p)" by auto
    from sol[unfolded match_decision] match_prob_of_rp_sound[OF _ mp] show ?r unfolding rp by force
  qed
qed


definition redex_rps_decision :: "('f, 'v) substL \<Rightarrow> ('f, 'v) redex_prob list \<Rightarrow> ('f, 'v) redex_prob check" where
  "redex_rps_decision \<mu> \<equiv> 
    let
      \<mu>' = subst_incr \<mu>;
      main = redex_decision \<mu>'
    in check_allm (\<lambda>tl. check (\<not> main tl) tl)"

lemma redex_rps_decision:
  "isOK (redex_rps_decision \<mu> rps) = (\<not>(\<exists>t l i C \<sigma>. (t, l) \<in> set rps \<and> t \<cdot> ((mk_subst Var \<mu>) ^^ i) = C\<langle>l \<cdot> \<sigma>\<rangle>))"
  (is "?l = ?r")
proof -
  interpret fixed_subst_incr "subst_incr \<mu>" .
  show ?thesis unfolding redex_rps_decision_def Let_def 
    unfolding redex_decision si_subst_subst_incr[symmetric]
    using redex_solution_def by force
qed

definition check_NF_iteration :: "('f, 'v) substL \<Rightarrow> ('f, 'v) term list \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) redex_prob check" where
  "check_NF_iteration \<mu> \<equiv> let dec = redex_rps_decision \<mu> in (\<lambda>Q t. dec (map (Pair t) Q))"

lemma check_NF_iteration:
  "isOK (check_NF_iteration \<mu> Q t) = (\<forall>i. t \<cdot> mk_subst Var \<mu> ^^ i \<in> NF_terms (set Q))"
  (is "?l = ?r") 
proof -
  have id: "?r = (\<not>(\<exists>q \<sigma> C i. q \<in> set Q \<and> t \<cdot> mk_subst Var \<mu> ^^ i = C\<langle>q \<cdot> \<sigma>\<rangle>))" 
    unfolding NF_ctxt_subst by auto
  show ?thesis unfolding id check_NF_iteration_def redex_rps_decision Let_def by force
qed

end

