(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory QDP_Framework_Impl
imports
  Dependency_Pair_Problem_Spec
  Termination_Problem_Spec
begin

type_synonym 'r result = "shows + 'r"
type_synonym 'a proc = "'a \<Rightarrow> 'a result"

definition (in tp_spec) sound_tt_impl :: "'a proc \<Rightarrow> bool" where
  "sound_tt_impl tt \<equiv> \<forall>tp tp'. tt tp = return tp' \<longrightarrow>
    SN_qrel (qreltrs tp') \<longrightarrow> SN_qrel (qreltrs tp)"

lemma (in tp_spec) sound_tt_implI[intro]:
  assumes "\<And>tp tp'. tt tp = return tp' \<Longrightarrow> SN_qrel (qreltrs tp') \<Longrightarrow> SN_qrel (qreltrs tp)"
  shows "sound_tt_impl tt"
  using assms unfolding sound_tt_impl_def by auto

lemma (in tp_spec) sound_tt_impl:
  assumes "sound_tt_impl tt"
  and "tt tp = return tp'"
  and "SN_qrel (qreltrs tp')"
  shows "SN_qrel (qreltrs tp)"
  using assms unfolding sound_tt_impl_def by auto

lemma (in dpp_spec) finite_under_var_cond_impl:
  assumes m: "M d"
  and nnfs: "\<not> NFS d"
  and var: "(\<And>lr. lr \<in> set (rules d) \<Longrightarrow> is_Fun (fst lr)) \<Longrightarrow> finite_dpp (dpp d)"
  shows "finite_dpp (dpp d)"
  using var unfolding rules_sound dpp_sound
  by (rule finite_under_var_cond[OF _ m nnfs])

definition (in dpp_spec) sound_proc_impl :: "'a proc \<Rightarrow> bool" where
  "sound_proc_impl proc \<equiv> \<forall>d d'. proc d = return d' \<longrightarrow>
    finite_dpp (dpp d') \<longrightarrow> finite_dpp (dpp d)"

lemma (in dpp_spec) sound_proc_implI[intro]:
  assumes "\<And>d d'. proc d = return d' \<Longrightarrow> finite_dpp (dpp d') \<Longrightarrow> finite_dpp (dpp d)"
  shows "sound_proc_impl proc"
  using assms unfolding sound_proc_impl_def by auto

lemma (in dpp_spec) sound_proc_impl:
  assumes "sound_proc_impl proc"
    and "proc d = return d'"
    and "finite_dpp (dpp d')"
  shows "finite_dpp (dpp d)"
  using assms unfolding sound_proc_impl_def by auto

lemma (in dpp_spec) rules_map_defined_conv:
  "(\<lambda>fn. rules_map d fn \<noteq> []) = defined (set (R d) \<union> set (Rw d))"
  by (rule ext) simp

definition internal_error where "internal_error \<equiv> (
  ''internal error'' +#+ shows_nl +@+
  ''please contact one of the CeTA members'' +#+ shows_nl +@+
  ''http://cl-informatik.uibk.ac.at/software/ceta'' +#+ shows_nl +@+
  ''ceta@informatik.uibk.ac.at'' +#+ shows_nl)"

definition
  shows_terms' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> string \<Rightarrow> ('f, 'v) term list \<Rightarrow> shows"
where
  "shows_terms' fun var name ts \<equiv>
    name +#+ shows_nl +@+ shows_list_gen (shows_term' fun var) [] [] [newline] [] ts +@+
    shows_nl"

definition
  shows_dpp' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> ('dpp, 'f, 'v) dpp_ops \<Rightarrow> 'dpp \<Rightarrow> shows"
where
  "shows_dpp' fun var I d = (let
    m = dpp_ops.minimal I d;
    nfs = dpp_ops.nfs I d;
    p  = dpp_ops.P  I d;
    pw = dpp_ops.Pw I d;
    r  = dpp_ops.R  I d;
    rw = dpp_ops.Rw I d;
    q  = dpp_ops.Q  I d
  in     
    shows_trs' fun var ''pairs:'' '' -> '' p +@+
    (if pw = [] then id else shows_trs' fun var ''weak pairs:'' '' ->= '' pw) +@+
    (if r  = [] then id else shows_trs' fun var ''strict rules:'' '' ->! '' r) +@+
    shows_trs' fun var ''rules:'' '' -> '' rw +@+
    (if q  = [] then id else shows_terms' fun var ''Q-component:'' q) +@+
    (if m then shows_nl +@+ shows ''(minimal)'' else id) +@+ 
    (if nfs \<and> q \<noteq> [] then shows_nl +@+ shows ''(normal form substitutions)'' else id))"

abbreviation "shows_dpp \<equiv> shows_dpp' shows shows"

definition
  shows_tp' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> ('tp, 'f, 'v) tp_ops \<Rightarrow> 'tp \<Rightarrow> shows"
where
  "shows_tp' fun var I t = (let
    nfs = tp_ops.nfs I t;
    r   = tp_ops.R  I t;
    rw  = tp_ops.Rw I t;
    q   = tp_ops.Q  I t
  in
    shows_trs' fun var ''rules:'' '' -> '' r +@+
    (if rw = [] then id else shows_trs' fun var ''relative rules:'' '' ->= '' rw) +@+
    (if q  = [] then id else shows_terms' fun var ''Q-component:'' q) +@+
    (if nfs then ''substitutions are assumed to be in normal form'' +#+ shows_nl else id))"

abbreviation "shows_tp \<equiv> shows_tp' shows shows"

definition indent :: "shows \<Rightarrow> shows" where
  "indent p s = concat (map (\<lambda>c. (if c = newline then newline # '' '' else [c])) (p s))"

end
