(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Tree_Automata_Checker
imports
  Xmlt
  Tree_Automata_Impl
  Linear_Orders
begin

(* TODO: several parts are just copied from CPF_Parser and
  these should be shared *)

fun
  var :: "('a, string) term xmlt"
where
  "var x = (Xmlt.text ''var'' x \<guillemotright>= (return \<circ> Var))"
hide_const (open) var

partial_function(sum_bot)
  "term" :: "'a xmlt \<Rightarrow> ('a, string) term xmlt"
where
  [code]: "term xml2name x = 
     Xmlt.options [(''var'', Tree_Automata_Checker.var), (''funapp'', Xmlt.many1 ''funapp'' xml2name (Xmlt.singleton ''arg'' (term xml2name) id) Fun)] x"  
hide_const (open) "term"

definition rule :: "'a xmlt \<Rightarrow> ('a, string) rule xmlt" where
  "rule xml2name = Xmlt.pair ''rule''
    (Xmlt.singleton ''lhs'' (Tree_Automata_Checker.term xml2name) id)
    (Xmlt.singleton ''rhs'' (Tree_Automata_Checker.term xml2name) id)
    Pair"
hide_const (open) rule 

definition rules :: "'a xmlt \<Rightarrow> ('a, string) rules xmlt" where
  "rules xml2name = Xmlt.many ''rules'' (Tree_Automata_Checker.rule xml2name) id"
hide_const (open) rules


definition state :: "string xmlt" where
  "state = Xmlt.text ''state''"


definition final_states :: "string list xmlt" where
  "final_states = Xmlt.many ''finalStates'' state id"

(* TODO: move this change to CPF_Parser (optional height) *)
definition transition_lhs :: "'a xmlt \<Rightarrow> ('a \<times> string list + string) xmlt" where
  "transition_lhs xml2name x = (if (case x of (XML _ _ cs _) \<Rightarrow> length cs = 1 \<and> (
                                    case hd cs of (XML name _ _ _) \<Rightarrow> name = ''state''
            )) then 
    (Xmlt.singleton ''lhs'' state Inr) x
   else
    Xmlt.many1 ''lhs''
    (xml2name)
    (state)
    (\<lambda>f qs. Inl (f, qs)) x)"

definition transition :: "'a xmlt \<Rightarrow> ((string, 'a) ta_rule + (string \<times> string)) xmlt" where
  "transition xml2name = Xmlt.pair ''transition''
    (transition_lhs xml2name)
    (Xmlt.singleton ''rhs'' state id)
    (\<lambda> lhs q. case lhs of Inl (f, qs) \<Rightarrow> Inl (TA_rule f qs q) | Inr q' \<Rightarrow> Inr (q',q))"

definition transitions :: "'a xmlt \<Rightarrow> (((string, 'a) ta_rule list) \<times> ((string \<times> string)list)) xmlt" where
  "transitions xml2name = Xmlt.many ''transitions'' (transition xml2name)
     (\<lambda> rls. let (rules, eps) = partition (\<lambda> rl. case rl of Inl _ \<Rightarrow> True | Inr _ \<Rightarrow> False) rls;
                  ruls = map (\<lambda> rl. case rl of Inl r \<Rightarrow> r) rules;
                  ep   = map (\<lambda> rl. case rl of Inr e \<Rightarrow> e) eps
                in (ruls,ep))"


definition tree_automaton :: "'a xmlt \<Rightarrow> (string, 'a) tree_automaton xmlt" where
  "tree_automaton xml2name \<equiv> Xmlt.pair ''treeAutomaton''
    final_states
    (transitions xml2name)
    (\<lambda> f (t, eps). Tree_Automaton f t eps)"

definition
  xml_tree_aut_problem ::
    "'a xmlt 
     \<Rightarrow> (string, 'a, string) tree_aut_problem xmlt"
where "xml_tree_aut_problem xml2name = Xmlt.triple ''treeAutomataProblem'' 
       (tree_automaton xml2name)
       (Xmlt.singleton ''trs'' (Tree_Automata_Checker.rules xml2name) id)
       (tree_automaton xml2name)
       Tree_Automata_Impl.tree_aut_problem"

primrec
  xmldoc2cert_problem :: "xmldoc \<Rightarrow> string +\<^sub>\<bottom> (string,string,string)tree_aut_problem"
where
  "xmldoc2cert_problem (XMLDOC header xml) = 
     xml_tree_aut_problem (Xmlt.text ''name'') xml"

(*include debug information for xml-parsing*)
definition doc_of_string :: "string \<Rightarrow> string + xmldoc" where
  "doc_of_string = debug ''0'' ''parsing txt to xml'' Xml.doc_of_string"

fun
  parse_cert_problem :: "string \<Rightarrow> string +\<^sub>\<bottom> (string,string,string) tree_aut_problem"
where
  "parse_cert_problem s = (case doc_of_string s of Inl x \<Rightarrow> error x | Inr y \<Rightarrow> xmldoc2cert_problem y)"


datatype cert_result = Certified | Unsupported string | Reject string 

fun certify_proof :: "string \<Rightarrow> cert_result"
where "certify_proof problemString  = (
  case parse_cert_problem problemString of
     Left err \<Rightarrow> Unsupported (''error while parsing'' @ [newline] @ err)
   | Right tap \<Rightarrow> case (certify_tree_aut_problem tap) of
                         Inr () \<Rightarrow> Certified 
                      |  Inl err  \<Rightarrow> Reject (err []))"

end



