{-
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
-}
module Main(main) where 

import Ceta -- the certifier
import System.Environment -- for getArgs
import System.IO -- for file reading
import System.Exit -- for error codes

main = do args <- getArgs
          case Prelude.length args of
            1 -> do problemString <- readFile (args !! 0)
                    start False problemString
            2 -> do let flag = args !! 0
                    if flag == "--allow-assumptions"
                      then do problemString <- readFile (args !! 1)
                              start True problemString
                      else error ("flag '" ++ flag ++ "' not supported")
            _ -> putStrLn "usage: ceta [--allow-assumptions] problemFile.xml(in cpf-format)     (CeTA version 2.12)"

start a problemString = 
    do case certify_proof a problemString of
         Sumbot (Inr (Certified prf)) -> 
             do putStrLn ("CERTIFIED " ++ prf)
            	exitSuccess
         Sumbot (Inr (Error message)) -> 
             do putStrLn "REJECTED"
                hPutStrLn stderr message
                exitWith (ExitFailure 1)
         Sumbot (Inr (Unsupported message)) -> 
             do putStrLn "UNSUPPORTED" 
                hPutStrLn stderr message
                exitWith (ExitFailure 2)
         Sumbot (Inl message) -> 
             do putStrLn "UNKNOWN ERROR"
                hPutStrLn stderr message
                exitWith (ExitFailure 3)


